# python3-opsi
# -*- coding: utf8 -*-
# This code is part of the opsi.org project
# http://opsi.org
#
# Copyright (c) uib gmbh (www.uib.de)
# This sourcecode is owned by the uib gmbh, D55118 Mainz, Germany
# and published under the Terms of the GNU Affero General Public License.
# Text of the AGPL: http://www.gnu.org/licenses/agpl-3.0-standalone.html

if not bootimageVersion or (int(bootimageVersion) < 20140118):
    raise Exception("This product requires opsi-linux-bootimage version >= 20140118")

try:
    if not backend.backend_info()['opsiVersion'].startswith('4.'):
        raise ValueError('No opsi 4.x backend present')
except Exception:
    raise Exception("This is a opsi 4.x product")

isOpsi43 = False
if backend.backend_info()['opsiVersion'].startswith('4.3'):
    isOpsi43 = True

try:
    import ConfigParser
except:
    import configparser
import subprocess
from OPSI.Util.File import IniFile
from OPSI.Backend.BackendManager import BackendManager
from OPSI.Backend.JSONRPC import JSONRPCBackend
from OPSI.Util import md5sum
from hashlib import md5
try:
    from twisted.conch.ssh import keys
except:
    pass
try:
    from urlparse import urlparse
except:
    from urllib.parse import urlparse

try:
    isWebdav
except NameError:
    isWebdav = False

opsiVersion = backend.backend_info()['opsiVersion']

#include opsi setup library
depot.copy(SCRIPT_PATH+'/opsisetuplib.py','/tmp')

exec(open('/tmp/opsisetuplib.py').read())

logger.notice(u"##############################################################")
initOpsiSetupLib()
logger.notice(u"nt6.py")
#logger.notice(u"nt6.py $Date$")
logger.notice(u"##############################################################")
productVersion =''
packageVersion =''
unattendTemplateMd5 = 'c307b3fa5f656441c1afa6e60602dadb'
for i in range(len(productOnDepots)):
    if (productOnDepots[i]['productId'] == productId):
        productVersion = productOnDepots[i]['productVersion']
        packageVersion = productOnDepots[i]['packageVersion']
    break
if isOpsi43:
    show_message(u"Installing Product: %s_%s-%s" % (productId,productVersion,packageVersion), 10)
logger.notice(u"Installing Product: %s_%s-%s on %s" % (productId,productVersion,packageVersion,clientId))
scriptMessageSubject.setMessage(u"Installing Product: %s_%s-%s on %s" % (productId,productVersion,packageVersion,clientId))
#logger.notice(u"Installing Product: %s on %s" % (productId,clientId))
logger.notice(u"##############################################################")
localbootProductsOnDepots = backend.productOnDepot_getIdents(productType = 'LocalbootProduct',depotId = depotId, returnType = 'dict')
scriptMessageSubject.setMessage(u"Check is opsi-script >= 4.12.0.13")
productVersion = '0.0.0.0'
winstProductVersion = '0.0.0.0'
for i in range(len(localbootProductsOnDepots)):
    aktprod = localbootProductsOnDepots[i]['productId']
    #logger.notice('Product found: %s' % aktprod)
    if aktprod == 'opsi-script':
        productVersion = localbootProductsOnDepots[i]['productVersion']
        action_processor = "opsi-script"
        break
    if aktprod == 'opsi-winst':
        winstProductVersion = localbootProductsOnDepots[i]['productVersion']
        action_processor = "opsi-winst"
if productVersion == '0.0.0.0':
    productVersion = winstProductVersion

logger.notice(u"Installed opsi-script version: %s " % (productVersion))
if compareVersions(productVersion,'<','4.12.0.13'):
    if isOpsi43:
        show_message(u"This product requires opsi-script >= 4.12.0.13")
    raise Exception("This product requires opsi-script >= 4.12.0.13")

try:
    if len(hostname) > 15:
        # With longer hostnames, the Windows installation will break shortly before finishing with a cryptic error:
        # Windows could not parse or process unattend answer file [C:\Windows\Panther\unattend.xml]
        # for pass [specialize]. The answer file is invalid.
        if isOpsi43:
            show_message(u"Hostname '{}' is too long. Maximal 15 characters allowed".format(hostname))
        raise Exception("Hostname '{}' is too long. Maximal 15 characters allowed".format(hostname))
except NameError:
    pass

inUefi = inUefiMode()
if inUefi:
    try:
        inSecureboot = isSecureboot()
    except NameError:
        inSecureboot = False

    if inSecureboot:
        logger.notice(u"we are running in Secure Boot mode")
        scriptMessageSubject.setMessage(u"we are running in Secure Boot mode")
    else:
        logger.notice(u"we are not running in Secure Boot mode")
        scriptMessageSubject.setMessage(u"we are not running in Secure Boot mode")


useGpt = False
if inUefi:
    useGpt = True

windows_firewall_enable = False
if (str(productProperties.get('windows_firewall_enable', 'True')).lower() == 'true'):
    windows_firewall_enable = True

system_keyboard_layout = productProperties.get('system_keyboard_layout', 'German=0407:00000407')
if '=' in system_keyboard_layout:
    system_keyboard_layout = system_keyboard_layout.split('=')[1]

winpe_inputlocale = productProperties.get('winpe_inputlocale', 'German=0407:00000407')
if '=' in winpe_inputlocale:
    winpe_inputlocale = winpe_inputlocale.split('=')[1]

installto  = productProperties.get('installto', 'disk')
if (installto == 'disk'):
    vhdmode = False
    diskmode = True
    olimode = False
    logger.notice(u"We install to disk")
    scriptMessageSubject.setMessage(u"We install to disk")

if (installto == 'vhd'):
    vhdmode = True
    diskmode = False
    olimode = False
    logger.notice(u"We install to VHD")
    scriptMessageSubject.setMessage(u"We install to VHD")

if (installto == 'oli'):
    vhdmode = False
    diskmode = False
    olimode = True
    logger.notice(u"We install to opsi-local-image")
    scriptMessageSubject.setMessage(u"We install to opsi-local-image")

if vhdmode or olimode:
    checkModules('local_imaging')

modifyPartitions = []
if useGpt:
    modifyPartitionTemplate = \
    '''
    <ModifyPartition wcm:action="add">
            <!-- Active>%(active)s</Active -->
            <Format>%(format)s</Format>
            <Label>%(label)s</Label>
            <Letter>%(letter)s</Letter>
            <!-- <Extend>false</Extend -->
            <Order>%(order)s</Order>
            <PartitionID>%(partition)s</PartitionID>
    </ModifyPartition>
    '''
else:
    modifyPartitionTemplate = \
    '''
    <ModifyPartition wcm:action="add">
            <Active>%(active)s</Active>
            <Format>%(format)s</Format>
            <Label>%(label)s</Label>
            <Letter>%(letter)s</Letter>
            <Extend>false</Extend>
            <Order>%(order)s</Order>
            <PartitionID>%(partition)s</PartitionID>
    </ModifyPartition>
    '''

target = '/mnt/hd'
source = SCRIPT_PATH
srcDriversDir = source + '/drivers'
dstDriversDir = target + '/drv'
dstPeDriversDir = target + '/drvpe'



bootPartitionNumber = 0
bootPartitionSizeM = 0
if vhdmode or olimode:
    bootPartitionSize   = "0"
    bootPartitionLabel  = ""
else:
    bootPartitionSize   = productProperties.get('boot_partition_size', '0')
    bootPartitionLabel  = productProperties.get('boot_partition_label', 'BOOT')

bootPartitionLetter = "S"
bootPartitionHidden = True
if diskmode:
    if (productProperties.get('boot_partition_letter', '-') != '-'):
        bootPartitionLetter = productProperties.get('boot_partition_letter')
        bootPartitionHidden = False

systemPartitionSizeM = 0
systemPartitionNumber = 1
if diskmode:
    windowsPartitionSize   = productProperties.get('windows_partition_size', '100%')
    systemPartitionLabel  = productProperties.get('windows_partition_label', 'WINDOWS')
    windowsPartitionLetter = "C"

if vhdmode:
    systemPartitionLabel  = 'CONTAINER'
    windowsPartitionLetter = "P"

if olimode:
    systemPartitionLabel  = 'system'
    windowsPartitionLetter = "C"


winpePartitionNumber = 2
winpePartitionSize   = "4000M"
winpePartitionLabel  = "WINPE"
winpePartitionLetter = "W"
if vhdmode:
    winpePartitionSize   = productProperties.get('winpe_partition_size', '4000M')
if diskmode:
    winpePartitionSize   = productProperties.get('winpe_partition_size', '4000M')

dataPartitionSizeM = 0
dataPartitionNumber = 0
dataPartitionLabel  = 'data'
dataPartitionLetter = 'D'
create_new_datapartition = False
dataPartitionCreate = False
if diskmode:
    dataPartitionLabel  = productProperties.get('data_partition_label', 'DATA')
    dataPartitionLetter = productProperties.get('data_partition_letter', 'D')
    dataPartitionCreate = True
    if (str(productProperties.get('data_partition_create', 'true')).lower() == 'false'):
        dataPartitionCreate = False

backupPartitionNumber = 3
backupPartitionCreate = False
backupPartitionLabel  = "opsidata"

bootimagePartitionSizeM = 0
bootimagePartitionSize = ""
bootimagePartitionLabel = "BOOTIMAGE"
bootimagePartitionNumber = 1
bootimagePartitionCreate = False

if vhdmode:
    windows_vhd_size = productProperties.get('windows_vhd_size', '70%')
    windows_vhd_sizeM = 0

blockAlignment = True

winpeNetworkMode = True
if (str(productProperties.get('winpenetworkmode', 'true')).lower() == 'false') or (isWebdav):
    winpeNetworkMode = False

force_del_winpe_boot_entry = False
if (str(productProperties.get('force_del_winpe_boot_entry', 'true')).lower() == 'true'):
    force_del_winpe_boot_entry = True

del_winpe_boot_entry_from_boot_sequence = False
if (str(productProperties.get('del_winpe_boot_entry_from_boot_sequence', 'true')).lower() == 'true'):
    del_winpe_boot_entry_from_boot_sequence = True

useRaid1 = False
if olimode or diskmode:
    if (str(productProperties.get('use_raid1', 'false')).lower() == 'true'):
        useRaid1 = True
logger.notice(u"useRaid1 is %s" % (str(useRaid1)))

preserve_winpe_partition = True
if diskmode:
    productProperties.get('windows_partition_size', '100%')
    logger.notice(u"preserve_winpe_partition is %s" % (str(productProperties.get('preserve_winpe_partition', ''))))
    if (str(productProperties.get('preserve_winpe_partition', 'false')).lower() == 'false'):
        preserve_winpe_partition = False

logger.notice(u"preserve_winpe_partition is %s" % (str(preserve_winpe_partition)))

installfiles_dir = 'installfiles'
if diskmode:
    try:
        installfiles_dir = productPropertyValues.get('installfiles_dir', [None])[0]
        if not installfiles_dir:
            installfiles_dir = 'installfiles'
    except IndexError:
        installfiles_dir = 'installfiles'
logger.notice(u"Using installfiles_dir: %s" % (installfiles_dir))

backupAfterInstall = True
if olimode:
    if (str(productProperties.get('backup_after_install', 'true')).lower() == 'false'):
        backupAfterInstall = False
        logger.notice(u"backupAfterInstall = False")
    else:
        logger.notice(u"backupAfterInstall = True")


multiDiskMode  = productProperties.get('multi_disk_mode', '0')

administratorPassword  = productProperties.get('administrator_password', 'nt123')
if administratorPassword == "":
    administratorPassword = "nt123"

winpeuilanguage = productProperties.get('winpe_uilanguage', 'en-US')

winpe_debug_cmd_exe = False
if (str(productProperties.get('winpe_debug_cmd_exe', 'false')).lower() == 'true'):
    winpe_debug_cmd_exe = True
    logger.notice(u"winpe_debug_cmd_exe = True")
else:
    logger.notice(u"winpe_debug_cmd_exe = False")

if olimode:
    # get multiDiskMode and backup_partition_on_same_disk from opsi-local-image-prepare
    multiDiskMode  = '0'
    backup_partition_on_same_disk = True
    dataPartitionSize = '0G'
    if hasattr(backend, "productPropertyState_getValues"):
        # opsi >= 4.3
        for property_id, values in (
            backend.productPropertyState_getValues(product_ids=['opsi-local-image-prepare'], object_ids=[clientId], with_defaults=True)[clientId]['opsi-local-image-prepare']
				.items()
			):
                logger.info("Found product property: %s=%r", property_id, values)
                if property_id == 'multi_disk_mode':
                    multiDiskMode  = values[0]
                if property_id == 'backup_partition_on_same_disk':
                    backup_partition_on_same_disk = values[0]
                if property_id == 'data_partition_size':
                    dataPartitionSize = values[0]
    else:
        backend.backend_setOptions( { 'addProductPropertyStateDefaults': True } )
        pps = backend.productPropertyState_getObjects(objectId = clientId, productId = 'opsi-local-image-prepare',propertyId = 'multi_disk_mode')
        if pps:
            for state in pps:
                myvalues = state.getValues()
                for key in myvalues:
                    multiDiskMode  = key
        pps = backend.productPropertyState_getObjects(objectId = clientId, productId = 'opsi-local-image-prepare',propertyId = 'backup_partition_on_same_disk')
        if pps:
            for state in pps:
                myvalues = state.getValues()
                for key in myvalues:
                    backup_partition_on_same_disk  = key
        pps = backend.productPropertyState_getObjects(objectId = clientId, productId = 'opsi-local-image-prepare',propertyId = 'data_partition_size')
        if pps:
            for state in pps:
                myvalues = state.getValues()
                for key in myvalues:
                    dataPartitionSize  = key
        logger.notice(u"multiDiskMode: %s ; backup_partition_on_same_disk: %s ; dataPartitionSize: %s" % (multiDiskMode,backup_partition_on_same_disk,dataPartitionSize))
    #pps = backend.productPropertyState_getObjwindows_vhd_sizewindows_vhd_sizeects(objectId = clientId, productId = 'opsi-local-image-prepare',propertyId = 'system_partition_size')
    #if pps:
    #       for state in pps:
    #               myvalues = state.getValues()
    #               for key in myvalues:
    #                       systemPartitionSize  = key
    #pps = backend.productPropertyState_getObjects(objectId = clientId, productId = 'opsi-local-image-prepare',propertyId = 'winpe_partition_size')
    #if pps:
    #       for state in pps:
    #               myvalues = state.getValues()
    #               for key in myvalues:
    #                       winpePartitionSize  = key
    # data
    dataPartitionSizeM = int(dataPartitionSize.upper().replace('M', '').replace('G', '').replace('B', ''))
    if (dataPartitionSize.upper().find('G') != -1):
        dataPartitionSizeM *= 1024
    if dataPartitionSizeM == 0:
        logger.notice(u"Required data partition size is %dM -> no data partition will be created" % (dataPartitionSizeM))
    else:
        dataPartitionNumber = 2
        winpePartitionNumber += 1
        backupPartitionNumber += 1
        logger.notice(u"Partition numbers are now: system=%d, data=%d, swap=%d, backup=%d" % (systemPartitionNumber, dataPartitionNumber, winpePartitionNumber, backupPartitionNumber))
        logger.notice(u"Required data partition (%d) size is %dM" % (dataPartitionNumber, dataPartitionSizeM))


logger.notice(u"Determine if we have to localize the administrator")
adminName = u'Administrator'
# we need to check the internal version of the setup.exe
if not depot.exists("%s/%s/setup.exe" % (source, installfiles_dir)):
    if isOpsi43:
        show_error("Expected setup.exe not found in %s/%s/" % (source, installfiles_dir))
    raise Exception("Expected setup.exe not found in %s/%s/" % (source, installfiles_dir))

# we can do this with the peres command (if available)
exitcode = os.system("which peres")
tempwinverlist = []
winverlist =[]
winver = '6'
depot.copy("%s/%s/setup.exe"  % (source, installfiles_dir), "/tmp/")
if (0 == int(exitcode)):
    logger.notice(u"peres found - we use peres -v to get setup.exe fileversion")
    tempwinverlist = execute("peres -v /tmp/setup.exe")
else:
    logger.notice(u"no peres found - we use getwinfileversion to get setup.exe fileversion")
    # get the correct architecture
    try:
        #result = execute("%s" % which(uname -p))
        result = execute("uname -p")
        if result:
            if result[0].strip() == "x86_64":
                arch = "64"
            else:
                arch = "32"
    except Exception as e:
        arch = "32"
        logger.error("Exception in uname -p")

    bootimgtarget = "/tmp"
    pathToCopy = os.path.join(source, "linutils", "%sbit" % arch,"getwinfileversion")
    depot.copy(pathToCopy, bootimgtarget)
    execute("chmod ugo+x %s/getwinfileversion" % (bootimgtarget))
    tempwinverlist = execute("%s/getwinfileversion /tmp/setup.exe" % (bootimgtarget))

if tempwinverlist:
    winverlist = filter(None, tempwinverlist)
    for winver in winverlist:
        winver = winver.strip()
        # remove leading 'Fileversion:    ' from peres
        if u':' in winver:
            winver = winver.split(':')[1]
            winver = winver.strip()
        # use only major number from something like 6.3.9.10
        if u'.' in winver:
            winver = winver.split('.')[0]
    logger.notice(u"winver is %s" % (winver))
# localize only if win version >= 10
if (int(winver) >= 10):
    logger.notice(u"we localize because winver %s is >= 10" % (winver))
    #https://social.technet.microsoft.com/wiki/contents/articles/13813.localized-names-for-administrator-account-in-windows.aspx
    #Language       Translation
    #Finnish        Järjestelmänvalvoja
    #French Administrateur
    #Hungarian      Rendszergazda
    #Portuguese (Brazil)    Administrador
    #Portuguese (Portugal)  Administrador
    #Russian        Администратор
    #Spanish        Administrador
    #Swedish        Administratör
    localized_adminList = {}
    localized_adminList['fi-FI'] =u'Järjestelmänvalvoja'
    localized_adminList['fr-FR'] =u'Administrateur'
    localized_adminList['hu-HU'] =u'Rendszergazda'
    localized_adminList['pt-BR'] =u'Administrador'
    localized_adminList['pt-PT'] =u'Administrador'
    localized_adminList['ru-RU'] =u'Администратор'
    localized_adminList['es-ES'] =u'Administrador'
    localized_adminList['sv-SE'] =u'Administratör'

    if winpeuilanguage in localized_adminList:
        adminName = localized_adminList[winpeuilanguage]

logger.notice('We use as administrator: %s' % adminName)


winpeDir = productProperties.get('winpe_dir', 'auto')
if winpeDir == 'auto':
    if inUefi:
        winpeDir = 'winpe_uefi'
    else:
        winpeDir = 'winpe'
if not depot.exists("%s/%s" % (source,winpeDir)):
    if isOpsi43:
        show_error("Expected WinPE Directory: %s not found in %s" % (winpeDir,source))
    raise Exception("Expected WinPE Directory: %s not found in %s" % (winpeDir,source))
if not depot.exists("%s/%s/sources/boot.wim" % (source,winpeDir)):
    if isOpsi43:
        show_error("Expected WinPE Directory: %s, but it seems to be empty" % (winpeDir))
    raise Exception("Found WinPE Directory: %s , but it seems to be empty" % (winpeDir))


# declare some variables that are needed even we are nor in winpenetworkmode
encodedPcpatchPassword = ""
shareusername = ""
hn = ""
sn = ""


if dataPartitionCreate and (windowsPartitionSize != "100%"):
    if olimode:
        # in oli we have boot,sys,data,pe (pe will never be deleted)
        dataPartitionNumber = 2
        winpePartitionNumber += 1
    else:
        # in disk we have boot,sys,pe,data (sys will grow if pe is deleted)
        dataPartitionNumber = 3
    create_new_datapartition = True

if not bootPartitionSize.startswith("0"):
    bootPartitionNumber = 1
    systemPartitionNumber += 1
    winpePartitionNumber += 1
    if (dataPartitionNumber > 0):
        dataPartitionNumber += 1

if inUefi:
    systemPartitionNumber += 2
    winpePartitionNumber += 2
    if (bootPartitionNumber > 0):
        bootPartitionNumber += 2
    if (dataPartitionNumber > 0):
        dataPartitionNumber += 2
    backupPartitionNumber += 2

if olimode:
    logger.notice(u"Partition numbers are now: system=%d, data=%s, swap=%d, backup=%d" % (systemPartitionNumber, dataPartitionNumber, winpePartitionNumber, backupPartitionNumber))
else:
    logger.notice(u"Partition numbers are now: , boot=%d, system=%d, winpe=%d, data=%s" % (bootPartitionNumber, systemPartitionNumber, winpePartitionNumber, dataPartitionNumber))
# check if it is allowed to delete the partition table
dataPartitionPreserve = 'never'
pre_format_system_partitions = True
if diskmode:
    if (str(productProperties.get('data_partition_preserve', 'never')).lower() == 'if_possible'):
        dataPartitionPreserve = 'if_possible'
    if (str(productProperties.get('data_partition_preserve', 'never')).lower() == 'always'):
        dataPartitionPreserve = 'always'

    pre_format_system_partitions = False
    if (str(productProperties.get('pre_format_system_partitions', 'false')).lower() == 'true'):
        pre_format_system_partitions = True

activePartition = "true"
if not bootPartitionSize.startswith("0"):
    modifyPartitions.append(
            modifyPartitionTemplate % {
                    "active":    activePartition,
                    "format":    "NTFS",
                    "label":     bootPartitionLabel,
                    "letter":    bootPartitionLetter.upper(),
                    "order":     len(modifyPartitions) + 1,
                    "partition": bootPartitionNumber,
            }
    )
    activePartition = "false"

if not vhdmode:
    # do not reformat the system partition in vhdmode - it will destroy the vhd
    modifyPartitions.append(
            modifyPartitionTemplate % {
                    "active":    activePartition,
                    "format":    "NTFS",
                    "label":     systemPartitionLabel,
                    "letter":    windowsPartitionLetter.upper(),
                    "order":     len(modifyPartitions) + 1,
                    "partition": systemPartitionNumber,
            }
    )

modifyPartitions.append(
        modifyPartitionTemplate % {
                "active":    "false",
                "format":    "",
                "label":     winpePartitionLabel,
                "letter":    winpePartitionLetter.upper(),
                "order":     len(modifyPartitions) + 1,
                "partition": winpePartitionNumber,
        }
)

if (dataPartitionNumber > 0):
    modifyPartitions.append(
            modifyPartitionTemplate % {
                    "active":    "false",
                    "format":    "",
                    "label":     dataPartitionLabel,
                    "letter":    dataPartitionLetter.upper(),
                    "order":     len(modifyPartitions) + 1,
                    "partition": dataPartitionNumber,
            }
    )
    

# moved down below we know the diskindex
#pi = open(PATCHA_IN, 'a')
#print >> pi, "windows_partition_number=%d" % systemPartitionNumber
#print >> pi, "winpe_partition_letter=%s" % winpePartitionLetter.upper()
#print >> pi, "modify_partitions=%s" % ''.join(modifyPartitions).replace('\n', '').replace('<Format></Format>', '')
#pi.close()

# Do hardware inventory
try:
    logger.notice(u"Fetching opsi hw audit configuration")
    hwconfig = backend.auditHardware_getConfig()
    if isOpsi43:
        show_message(u"Running hardware inventory", 10)
    logger.notice(u"Running hardware inventory")
    auditHardwareOnHosts = auditHardware(config = hwconfig, hostId = clientId)

    logger.notice(u"Sending hardware information to service")
    backend.auditHardwareOnHost_setObsolete(clientId)
    backend.auditHardwareOnHost_createObjects(auditHardwareOnHosts)
except Exception as AuditError:
    logger.warning("Something went wrong during Hardware Audit, driver integration might not be complete")
    logger.error("%s" % AuditError)
    pass

if isOpsi43:
    show_message("Preparing disk", 20)
logger.notice(u"Swapoff all to avoid disk in use .....")
execute("swapoff -a -v")
lvmClearVolumeGroups()
# Get harddisks
disks = getHarddisks()
disks = sorted(disks, key=lambda disk: disk.device)
showDiskOrder(disks)
logger.notice(u"Normally we using disk 0")
diskindex = getDiskIndexFromMultiDiskMode(multiDiskMode)
installdiskindex = diskindex
logger.notice(u"multi_disk_mode is %s - so we use disk number: %d uuid:%s" %(multiDiskMode,diskindex,getUuidFromDiskDevice(disks[diskindex].device)))
disk = disks[diskindex]
logger.notice(u"Using disk device %s" % (disk.device))
dummy = isRotationalDisk(disk.device)

if olimode:
    backupdiskindex, backup_partition_on_same_disk = getBackupdiskindexFromBackupPartitionOnSameDisk(backup_partition_on_same_disk)
    if (backup_partition_on_same_disk == False):
        backupPartitionNumber = 1

if vhdmode:
    installdiskindex = len(disks)

# Calculate partitions
disk.setBlockAlignment(blockAlignment)
diskSizeM = (float(disk.size)/(1024*1024))
diskTotalSectors = disk.sectors


if diskmode or vhdmode:
    if not bootPartitionSize.startswith("0"):
    # Create boot partition
        bootPartitionSizeM = int(bootPartitionSize.upper().replace('M', '').replace('G', '').replace('B', ''))
        if (bootPartitionSize.upper().find('G') != -1): 
            bootPartitionSizeM *= 1024
            logger.notice(u"Required Boot partition (%d) size is %dM" % (bootPartitionNumber, bootPartitionSizeM))

    winpePartitionSizeM = int(winpePartitionSize.upper().replace('M', '').replace('G', '').replace('B', ''))
    if (winpePartitionSize.upper().find('G') != -1):
        winpePartitionSizeM *= 1024
    logger.notice(u"Required winpe partition (%d) size is %dM" % (winpePartitionNumber, winpePartitionSizeM))

    peStartM = int(diskSizeM - winpePartitionSizeM)
    peEndM = int(diskSizeM)
    systemPartitionSizeM = int(diskSizeM - bootPartitionSizeM - winpePartitionSizeM - 1000)
    logger.notice(u"systemPartitionSizeM is: %d" % systemPartitionSizeM)
    boundaryM = int(diskSizeM)
    if vhdmode:
        if (windows_vhd_size == "100%"):
            windows_vhd_sizeM = int((float(80)/100.0)*float(systemPartitionSizeM))
            logger.notice(u"windows_vhd_size is set from 100 percent to 80 percent of disk = %dM" % windows_vhd_sizeM)
            logger.notice(u"We now have %s GB freespace for the child VHD" % (int(float((diskSizeM - winpePartitionSizeM - windows_vhd_sizeM - 1000) / 1024))))
            if int(systemPartitionSizeM - windows_vhd_sizeM) < int(25600):
                windows_vhd_sizeM = int(systemPartitionSizeM - 26000)
                logger.warning("windows_vhd_size is set to %dM in order to have 25GB freespace for the child VHD" % windows_vhd_sizeM)
        else:
            windows_vhd_sizeM = 0
            if (windows_vhd_size.find('%') != -1):
                percent = int(windows_vhd_size.replace('%', ''))
                windows_vhd_sizeM = int((float(percent)/100.0)*float(systemPartitionSizeM))
            else:
                windows_vhd_sizeM = int(windows_vhd_size.upper().replace('M', '').replace('G', '').replace('B', ''))
                if (windows_vhd_size.upper().find('G') != -1):
                    windows_vhd_sizeM *= 1024
            if int(systemPartitionSizeM - windows_vhd_sizeM) < int(25600):
                windows_vhd_sizeM = int(systemPartitionSizeM - 26000)
                logger.warning("windows_vhd_size is set to %dM in order to have 25GB freespace for the child VHD" % windows_vhd_sizeM)
            logger.notice(u"Required windows_vhd_size is %dM" % (windows_vhd_sizeM))
            logger.notice(u"Sizes: disk %dG, system %dG, winpe %dG" % (diskSizeM / 1024, systemPartitionSizeM / 1024, winpePartitionSizeM / 1024))
    else:
        if (windowsPartitionSize != "100%"):
            systemPartitionSizeM = 0
            if (windowsPartitionSize.find('%') != -1):
                percent = int(windowsPartitionSize.replace('%', ''))
                systemPartitionSizeM = int((float(percent)/100.0)*float(diskSizeM - bootPartitionSizeM - winpePartitionSizeM))
            else:
                systemPartitionSizeM = int(windowsPartitionSize.upper().replace('M', '').replace('G', '').replace('B', ''))
                if (windowsPartitionSize.upper().find('G') != -1):
                    systemPartitionSizeM *= 1024
            systemPartitionSizeM = int(systemPartitionSizeM - 1000)
            dataPartitionSizeM = diskSizeM - systemPartitionSizeM - bootPartitionSizeM - winpePartitionSizeM - 1000
            logger.notice(u"Required Windows partition (%s) size is %dM, data partition (%d) size is %dM" % (systemPartitionNumber, systemPartitionSizeM, dataPartitionNumber, dataPartitionSizeM))
            if (dataPartitionSizeM <= 0):
                if isOpsi43:
                    show_error(u"Not enough disk space for required windows partition size %dM" % windowsPartitionSize)
                raise Exception("Not enough disk space for required windows partition size %dM" % windowsPartitionSize)
            boundaryM = bootPartitionSizeM + systemPartitionSizeM
            peStartM = bootPartitionSizeM + systemPartitionSizeM
            peEndM = peStartM + winpePartitionSizeM

    delete_winpe = False
    # should we preserve the partition table ?
    if (dataPartitionPreserve == 'never'):
        will_create_new_partition_table = True
    else:
        will_create_new_partition_table = False

        logger.notice(u"Check if preserve data partition is possible ....")
        # Get current partitions
        partitions = disk.getPartitions()

        if not partitions:
            # No partition found on harddisk
            scriptMessageSubject.setMessage(u"Found no partitions on %s " % disk.device)
            logger.notice(u"Found no partitions on %s " % disk.device)
            # nothing to preserve
            will_create_new_partition_table = True
        else:
            if (windowsPartitionSize == "100%"):
                scriptMessageSubject.setMessage(u"Ignoring dataPartitionPreserve = %s because we have a required windowsPartitionSize = 100 percent" % dataPartitionPreserve)
                logger.notice(u"Ignoring dataPartitionPreserve = %s because we have a required windowsPartitionSize = 100 percent" % dataPartitionPreserve)
                will_create_new_partition_table = True
                dataPartitionPreserve = 'never'
            else:
                # we have a partition table and we should preserve it
                # is it possible ?
                if (len(partitions) < systemPartitionNumber):
                    logger.notice(u"Ignoring dataPartitionPreserve because number of existing partitions %s is lower than the required system partition number %s." % (len(partitions), systemPartitionNumber))
                    will_create_new_partition_table = True
                else:
                    if not useGpt:
                        logger.notice(u"We do not know if disk %s has a usable legacy partition table" % (disk.device))
                    else:
                        logger.notice(u"has disk %s a usable gpt partition table ?" % (disk.device))
                        fdiskexitcode = os.system("sgdisk -p %s >> /tmp/gpt.log 2>&1" % (disk.device))
                        if (0 != int(fdiskexitcode)):
                            logger.notice(u"disk %s have no usable gpt partition table" % (disk.device))
                            scriptMessageSubject.setMessage(u"disk %s have no usable gpt partition table" % (disk.device))
                            will_create_new_partition_table = True
                        else:
                            logger.notice(u"disk %s have a usable gpt partition table:" % (disk.device))
                            execute("sgdisk -p %s" % (disk.device))

                    # check the existing partitions
                    if not bootPartitionSize.startswith("0"):
                        # we should have a bootpartion
                        if useGpt:
                            logger.notice(u"Check boot partition size ...")
                            parsizes = getGptPartitionSizesInMegabytes(disk.device, bootPartitionNumber)
                            if parsizes:
                                SizeP = parsizes[bootPartitionNumber]
                            else:
                                logger.notice(u"boot partition (%d) seems not to exist" % (bootPartitionNumber))
                                SizeP = 0
                        else:
                            try:
                                SizeP = int(float(disk.getPartition(bootPartitionNumber).get("size"))/(1024*1024))
                                logger.notice(u"Existing boot partition size = %s" % SizeP)
                            except:
                                logger.notice(u"boot partition (%d) seems not to exist" % (bootPartitionNumber))
                                SizeP = 0

                        # tolerance 50 M
                        if (abs(bootPartitionSizeM - SizeP) > 50):
                            logger.notice(u"Existing boot partition size %s does not match on required size %dM" % (SizeP, bootPartitionSizeM))
                            scriptMessageSubject.setMessage(u"Existing boot partition size %s does not match on required size %dM" % (SizeP, bootPartitionSizeM))
                            will_create_new_partition_table = True

                    # check windows partition
                    freespaceM = 0
                    if useGpt:
                        logger.notice(u"Check system partition size ...")
                        parsizes = getGptPartitionSizesInMegabytes(disk.device, systemPartitionNumber)
                        if parsizes:
                            SizeP = parsizes[systemPartitionNumber]
                        else:
                            logger.notice(u"system partition (%d) seems not to exist" % (systemPartitionNumber))
                            SizeP = 0
                        ###
                    else:
                        try:
                            SizeP = int(float(disk.getPartition(systemPartitionNumber).get("size"))/(1024*1024))
                            logger.notice(u"Existing windows partition size = %s" % SizeP)
                        except:
                            logger.notice(u"windows partition (%d) seems not to exist" % (systemPartitionNumber))
                            SizeP = 0
                            try:
                                freespaceM = int(partitions[1]['end'] / (1024*1024))
                                logger.notice(u"We have a free space of %sM before partition 2" % (freespaceM))
                            except:
                                pass
                    SizePe = 0

                    exitcode = os.system("test -b %s >> /dev/null 2>&1" % (getPartitionDevicePath(disk.device, winpePartitionNumber)))
                    if (0 == int(exitcode)):
                        logger.notice(u"%s exists" % (getPartitionDevicePath(disk.device, winpePartitionNumber)))
                        partitionLabel = 'unknown'
                        resultlist = execute('blkid -o value -s LABEL %s' % (getPartitionDevicePath(disk.device, winpePartitionNumber)))
                        resultlist = list(filter(None, resultlist))
                        for resultline in resultlist:
                            try:
                                partitionLabel = resultline.strip()
                            except:
                                pass
                        logger.notice(u"%s exists as device with label: %s" % (getPartitionDevicePath(disk.device, winpePartitionNumber), partitionLabel) )
                        if (partitionLabel == winpePartitionLabel):
                            logger.notice(u"%s exists unexpected as winpe partition- perhaps reusing it" % (getPartitionDevicePath(disk.device, winpePartitionNumber)))
                            if useGpt:
                                logger.notice(u"Check winpe partition size ...")
                                parsizes = getGptPartitionSizesInMegabytes(disk.device, winpePartitionNumber)
                                if parsizes:
                                    SizePe = parsizes[winpePartitionNumber]
                                else:
                                    logger.notice(u"winpe partition (%d) seems not to exist" % (winpePartitionNumber))
                                    SizePe = 0
                                ###
                            else:
                                SizePe = int(float(disk.getPartition(winpePartitionNumber).get("size"))/(1024*1024))
                            logger.notice(u"Existing winpe partition size = %s" % SizePe)
                            if (abs(winpePartitionSizeM - SizePe) > 10):
                                logger.notice(u"Existing winpe partition size %s does not match the required partion size %s - the partition will be deleted." % (SizePe, winpePartitionSizeM))
                                delete_winpe = True
                            else:
                                logger.notice(u"Existing winpe partition size %s does match the required partion size %s - the partition will be reused." % (SizePe, winpePartitionSizeM))
                                delete_winpe = False

                        else:
                            logger.notice(u"No winpe partition found (as expected)")
                    # tolerance 5G
                    if (((SizeP + SizePe + freespaceM) - (systemPartitionSizeM)) < 0):
                        logger.notice(u"Existing free space, windows (and PE) partition size %d does not match the required size %dM" % ((SizeP+SizePe+freespaceM), systemPartitionSizeM))
                        will_create_new_partition_table = True
                    else:
                        logger.notice(u"Existing free space windows (and PE) partition size %d match the required size %dM" % ((SizeP+SizePe+freespaceM), systemPartitionSizeM))

        if will_create_new_partition_table:
            if (dataPartitionPreserve == 'always'):
                # we have to stop here
                scriptMessageSubject.setMessage(u"Could not use the existing partition table on %s - but not allowed to delete it - stopping here" % (disk.device))
                if isOpsi43:
                    show_error("Could not use the existing partition table on %s - but not allowed to delete it - stopping here" % (disk.device))
                raise Exception("Could not use the existing partition table on %s - but not allowed to delete it - stopping here" % (disk.device))
            else:
                scriptMessageSubject.setMessage(u"Could not use the existing partition table on %s - but allowed to delete it - going on" % (disk.device))
        else:
            scriptMessageSubject.setMessage(u"Will use the existing partition table on %s - going on" % (disk.device))

if olimode:
    checkOliPartitions(diskindex,backupdiskindex,useGpt,backup_partition_on_same_disk)
    disk = disks[diskindex]
    # change partition id and boot flag
    if inUefi:
        execute('efibootmgr')
        #system
        logger.notice(u"change id of partition %s to 0700" % getPartitionDevicePath(disk.device,systemPartitionNumber))
        execute('/usr/sbin/sgdisk -t %d:0700 --attributes %d:set:63 --attributes %d:clear:0 -p %s' % (systemPartitionNumber, systemPartitionNumber, systemPartitionNumber, disk.device))
        execute('/usr/sbin/sgdisk --info=%d %s' % (systemPartitionNumber, disk.device))
        #execute('sfdisk --change-id %s %d 07' % (disk.device,systemPartitionNumber))
        #winpe/swap
        logger.notice(u"change id of partition %s to 0700" % getPartitionDevicePath(disk.device,winpePartitionNumber))
        execute('/usr/sbin/sgdisk -t %d:0700 --attributes %d:clear:62 -p %s' % (winpePartitionNumber, winpePartitionNumber, disk.device))
        execute('/usr/sbin/sgdisk --info=%d %s' % (winpePartitionNumber, disk.device))
        #data
        if (dataPartitionNumber >= 2):
            logger.notice(u"change id of partition %s%d to 0700" % (disk.device,dataPartitionNumber))
            execute('/usr/sbin/sgdisk -t %d:0700 --attributes %d:set:63 --attributes %d:clear:0 -p %s' % (dataPartitionNumber, dataPartitionNumber, dataPartitionNumber, disk.device))
            execute('/usr/sbin/sgdisk --info=%d %s' % (dataPartitionNumber, disk.device))

        deleteAllEntriesFromEFIBootmgr('opsitempwinpe')
        deleteAllEntriesFromEFIBootmgr('Windows Boot Manager')
        deleteAllEntriesFromEFIBootmgr('opsibootimage')
        deleteAllEntriesFromEFIBootmgr('installer')
        deleteAllEntriesFromEFIBootmgr('ubuntu')
        logger.notice(u"setting winpe to the top of winpe boot order")
        try:
            invalidArgument = False
            execute('/bin/efibootmgr -c --disk %s --part %s -L opsitempwinpe --loader \\\\EFI\\\\boot\\\\bootx64.efi' % (disk.device, winpePartitionNumber))
        except Exception as efibootmgrError:
            logger.error("efibootmgrError")
            logger.error("%s" % efibootmgrError)
            if "Invalid argument" in efibootmgrError.args[0]:
                logger.error("Something went wrong with efibootmgr 'Invalid argument'. Working around.")
                invalidArgument = True
                pass
        if not invalidArgument:
            setFirstUefiBoot('opsitempwinpe')
            setNextUefiBoot('opsitempwinpe')
        execute('/bin/efibootmgr -v')
    else:
        logger.notice(u"change id of partition %s to 07" % getPartitionDevicePath(disk.device,systemPartitionNumber))
        execute('sfdisk --change-id %s %d 07' % (disk.device,systemPartitionNumber))
        logger.notice(u"change id of partition %s to 0b" % getPartitionDevicePath(disk.device,winpePartitionNumber))
        execute('sfdisk --change-id %s %d 0b' % (disk.device,winpePartitionNumber))
        logger.notice(u"Activate (set boot flag) on partition %s" % getPartitionDevicePath(disk.device,winpePartitionNumber))
        execute('sfdisk -A %s %d' % (disk.device,winpePartitionNumber))
        exitcode = os.system('%s %s >> /dev/null 2>&1' % (which('partprobe'), disk.device))
        execute('sfdisk -l')
        logger.notice(u"Write Master Boot Record")
        disk.writeMasterBootRecord(system = 'vista')
        #logger.notice(u"Write Partition Boot Record")
        #disk.writePartitionBootRecord(partition = winpePartitionNumber, fsType = 'fat32nt60')


# create mount point before loop
checkCreateDir(target)

unusedPartitionIdBackup = []
maxDiskRange = int(diskindex+1)
if (len(disks) > 1):
    if useRaid1:
        diskindex = 0
        maxDiskRange = int(2)
    else:
        if (multiDiskMode == "0_ignore_anything_else"):
            diskindex = 0
            maxDiskRange = int(1)
        else:
            logger.notice(u"We hide all partitions ....")
            for i in range(len(disks)):
                aktdisk = disks[i]
                if (i != diskindex) and not isRaidMember(aktdisk.device):
                    execute('sfdisk -l %s' % (aktdisk.device))
                    logger.notice(u"has disk %s a gpt partition table ?" % (aktdisk.device))
                    fdiskexitcode = os.system('blkid -s PTTYPE %s | grep -qF gpt' % (aktdisk.device))
                    if (0 != int(fdiskexitcode)):
                        logger.notice(u"disk %s have no gpt partition table" % (aktdisk.device))
                        aktDiskIsGPT = False
                        numberOfPartitions = len(aktdisk.getPartitions())
                    else:
                        logger.notice(u"disk %s have a gpt partition table:" % (aktdisk.device))
                        aktDiskIsGPT = True
                        #output = execute('parted -s -m %s print' % (aktdisk.device))
                        #output = output[len(output)-1].split(':')
                        #numberOfPartitions = int(output[0])
                        output = execute('sgdisk -p %s' % (aktdisk.device))
                        output = list(filter(None, output))
                        output = output[len(output)-1].split()
                        if output[0] == 'Number':
                            numberOfPartitions = 0
                        else:
                            numberOfPartitions = int(output[0])
                    logger.notice(u"We have %d partitions on disk %s" % (numberOfPartitions,aktdisk.device))
                    for aktpartnum in range(1,numberOfPartitions+1):
                        #aktpart =
                        if aktDiskIsGPT:
                            output = execute('/usr/sbin/sgdisk --info=%d %s | grep "Attribute flags"' % (aktpartnum, aktdisk.device))
                            output = output[0].split(':')
                            oldPartitionId = output[1]
                            oldPartitionId = oldPartitionId.strip()
                            unusedPartitionIdBackup.append([i,aktpartnum,aktDiskIsGPT,oldPartitionId])
                            logger.notice(u"Old Values: Disk: %d, part: %d, id: %s" % (i,aktpartnum,oldPartitionId))
                            logger.notice(u"change id of partition %s to 0700 / 63" % (getPartitionDevicePath(aktdisk.device,aktpartnum)))
                            execute('/usr/sbin/sgdisk -t %d:0700 --attributes %d:set:63 -p %s' % (aktpartnum, aktpartnum, aktdisk.device))
                            execute('/usr/sbin/sgdisk --info=%d %s' % (aktpartnum, aktdisk.device))
                        else:
                            output = execute('sfdisk --print-id %s %d' % (aktdisk.device,aktpartnum))
                            oldPartitionId = output[0].strip()
                            unusedPartitionIdBackup.append([i,aktpartnum,aktDiskIsGPT,oldPartitionId])
                            logger.notice(u"Old Values: Disk: %d, part: %d, id: %s" % (i,aktpartnum,oldPartitionId))
                            logger.notice(u"change id of partition %s to 12" % (getPartitionDevicePath(aktdisk.device,aktpartnum)))
                            execute('sfdisk --change-id %s %d 12' % (aktdisk.device,aktpartnum))
                    execute('%s %s' % (which('partprobe'),aktdisk.device))
                    time.sleep(2)
                    partitions = aktdisk.getPartitions()
                    execute('sfdisk -l %s' % (aktdisk.device))

win11_hardware_check = "false"
if (str(productProperties.get('win11_hardware_check', 'False')).lower() == 'true'):
    win11_hardware_check = "true"

with open(PATCHA_IN, 'a') as pi:
    pi.write("adminName=%s\n" % adminName)
    pi.write("diskindex=%d\n" % diskindex)
    # installdiskindex for vhd unattend = number of real disks + 1 (the vhd) -1 (to get the index starting with 0)
    # installdiskindex is only different to diskindex at opsi-vhd products
    pi.write("installdiskindex=%d\n" % installdiskindex)
    pi.write("administratorPassword=%s\n" % administratorPassword)
    if vhdmode:
        if useGpt:
            pi.write("windows_partition_number=3\n")
        else:
            pi.write("windows_partition_number=1\n")
    else:
        pi.write("windows_partition_number=%d\n" % systemPartitionNumber)
    pi.write("winpe_partition_letter=%s\n" % winpePartitionLetter.upper())
    pi.write("modify_partitions=%s\n" % ''.join(modifyPartitions).replace('\n', '').replace('<Format></Format>', ''))
    pi.write("system_keyboard_layout=%s\n" % system_keyboard_layout)
    pi.write("winpe_inputlocale=%s\n" % winpe_inputlocale)
    pi.write("windows_firewall_enable=%s\n" % windows_firewall_enable)
    pi.write("win11_hardware_check=%s\n" % win11_hardware_check)

logger.notice(u"maxDiskRange is %d" % (maxDiskRange))

for diskcounter in range(diskindex,maxDiskRange):
    logger.notice(u"diskcounter is now %d" % (diskcounter))
    disk = disks[diskcounter]
    showDiskOrder(disks)
    logger.notice(u"Using disk device %s" % (disk.device))
    dummy = isRotationalDisk(disk.device)

    disk.setBlockAlignment(blockAlignment)
    scriptMessageSubject.setMessage(u"Using harddisk %s (%0.0f MB)." % (disk.device, (float(disk.size)/(1024*1024))))

    # Check disk size
    if (disk.size < 18000*1024*1024):
        # Disk smaller than 18000 MB => give up
        if isOpsi43:
            show_error("Error: Disk too small")
        raise Exception("Disk too small")

    # Get current partitions
    partitions = disk.getPartitions()

    if not partitions:
        # No partition found on harddisk
        scriptMessageSubject.setMessage(u"Found no partitions on %s" % disk.device)

    if not olimode:
        if will_create_new_partition_table:
            # Create new partition table
            scriptMessageSubject.setMessage(u"Create new partition table on %s." % (disk.device))
            logger.notice(u"Clear first 512 bytes on %s to remove grub and so on ..." % (disk.device))
            fdiskexitcode = os.system('dd if=/dev/zero of=%s bs=512 count=10 >> /dev/null 2>&1' % (disk.device))
            execute('%s --rereadpt %s' % (which('blockdev'),disk.device))
            time.sleep(2)
            if useGpt:
                logger.notice(u"Clear partition table of %s" % (disk.device))
                fdiskexitcode = os.system('/usr/sbin/sgdisk --zap-all %s >> /dev/null 2>&1' % (disk.device))
                execute('/usr/sbin/sgdisk --clear -g -p %s' % (disk.device))
                if inUefi:
                    logger.notice(u"Create EFI system partition on %s" % (getPartitionDevicePath(disk.device,1)))
                    execute('/usr/sbin/sgdisk -g --new 1::400M -t 1:ef00 -c 1:\"EFI system partition\" --partition-guid=C12A7328-F81F-11D2-BA4B-00A0C93EC93B -p %s' % (disk.device))
                    execute('/bin/sleep 2')
                    execute('/usr/sbin/sgdisk -p %s' % (disk.device))
                    logger.notice(u"Make file system on EFI system partition on %s" % (getPartitionDevicePath(disk.device,1)))
                    execute('/sbin/mkfs.vfat %s' % (getPartitionDevicePath(disk.device,1)))
                    logger.notice(u"Create MS reserved partition on %s" % (getPartitionDevicePath(disk.device,2)))
                    execute('/usr/sbin/sgdisk -g --new 2::+128M -t 2:0c01 -c 2:\"MS reserved\" --partition-guid=E3C9E316-0B5C-4DB8-817D-F92DF00215AE -p %s' % (disk.device))
            else:
                logger.notice(u"using sgdisk to clear not usable partition table of %s" % (disk.device))
                fdiskexitcode = os.system('/usr/sbin/sgdisk --zap-all %s >> /dev/null 2>&1' % (disk.device))
                fdiskexitcode = os.system("echo %s1 : start=0, size=0 | sfdisk -L --no-reread %s >> /dev/null 2>&1" % (disk.device,disk.device))
                scriptMessageSubject.setMessage(u"Deleted partition Table on %s - Perhaps you need a restart....." % (disk.device))
                disk.deletePartitionTable()
                # Write Master Boot Record
                #disk.writeMasterBootRecord(system = 'vista')
                logger.notice(u"Write Master Boot Record")
                execute("/usr/local/bin/ms-sys -v")
                if int(bootimageVersion) < 20181022:
                    execute("/usr/local/bin/ms-sys --mbrnt60 %s" % (disk.device))
                else:
                    execute("/usr/local/bin/ms-sys --mbr7 %s" % (disk.device))
                    # That can not be done here: no partitions right now:
                    #execute("/usr/local/bin/ms-sys --ntfs %s" % (getPartitionDevicePath(disk.device, winpePartitionNumber)))
        else:
            logger.notice(u"we reuse the existing partitiontable")
            if useGpt:
                if not bootPartitionSize.startswith("0"):
                    logger.notice(u"Delete Partitions ...")
                    execute('/usr/sbin/sgdisk -g --delete=%d -p %s' % (bootPartitionNumber, disk.device))
                    execute('/usr/sbin/sgdisk -g --delete=%d -p %s' % (systemPartitionNumber, disk.device))
                else:
                    logger.notice(u"Delete Partitions ...")
                    execute('/usr/sbin/sgdisk -g --delete=%d -p %s' % (systemPartitionNumber, disk.device))
            else:
                if not bootPartitionSize.startswith("0"):
                    # boot
                    execute('sfdisk --delete %s %s' % (disk.device, bootPartitionNumber))
                    # system
                    execute('sfdisk --delete %s %s' % (disk.device, systemPartitionNumber))
                else:
                    # system
                    execute('sfdisk --delete %s %s' % (disk.device, systemPartitionNumber))
                if delete_winpe:
                    execute('sfdisk --delete %s %s' % (disk.device, winpePartitionNumber))

        # Partition table deleted => delete all netboot product states
        deletePocs = []
        for poc in backend.productOnClient_getObjects(clientId = clientId):
            if ((poc.productType == 'NetbootProduct') and (poc.productId != productId)) or \
               ((poc.productType == 'LocalbootProduct') and (poc.installationStatus == 'not_installed') and (poc.actionRequest == 'none')):
                deletePocs.append(poc)
        backend.productOnClient_deleteObjects(deletePocs)

        # Create partitions
        if not bootPartitionSize.startswith("0"):
            createPartitionEx("boot", bootPartitionNumber, bootPartitionLabel, 0, int(bootPartitionSizeM), gptAutoAppend = True, boot = False, gptAttribute = "63" , useGpt = useGpt, diskDevice = disk.device, fileSystem = "ntfs", gptFileSystem="0700")
            if pre_format_system_partitions:
                createFilesystemEx("boot", bootPartitionNumber, bootPartitionLabel, "ntfs", disk.device, useGpt)
            if bootPartitionNumber and bootPartitionHidden:
                if inUefi:
                    logger.notice(u"change id of partition %s to 0700 / 63" % (getPartitionDevicePath(disk.device,bootPartitionNumber)))
                    execute('/usr/sbin/sgdisk -t %d:0700 --attributes %d:set:63 -p %s' % (bootPartitionNumber, bootPartitionNumber, disk.device))
                    execute('/usr/sbin/sgdisk --info=%d %s' % (bootPartitionNumber, disk.device))
                else:
                    logger.notice(u"change id of partition %s to 12" % (getPartitionDevicePath(disk.device,bootPartitionNumber)))
                    execute('sfdisk --change-id %s %d 12' % (disk.device,bootPartitionNumber))
                execute('%s %s' % (which('partprobe'),disk.device))
                time.sleep(2)
                partitions = disk.getPartitions()
                execute('sfdisk -l %s' % (disk.device))
        # Create system partition
        createPartitionEx("system", systemPartitionNumber, systemPartitionLabel, 0, systemPartitionSizeM, gptAutoAppend = True, boot = False, gptAttribute = 63 , useGpt = useGpt, diskDevice = disk.device, fileSystem = "ntfs", gptFileSystem="0700")
        if inUefi:
            #logger.notice(u"change id of partition %s to 0b00 / 63" % (getPartitionDevicePath(disk.device,systemPartitionNumber)))
            #execute('/usr/sbin/sgdisk -t %d:0b00 --attributes %d:set:63 -p %s' % (systemPartitionNumber, systemPartitionNumber, disk.device))
            execute('/usr/sbin/sgdisk --info=%d %s' % (systemPartitionNumber, disk.device))
        if pre_format_system_partitions:
            createFilesystemEx("system", systemPartitionNumber, systemPartitionLabel, "ntfs", disk.device, useGpt)

        if not useGpt:
            partitions = disk.getPartitions()

        # Create winpe partition
        logger.notice(u"Check winpe partition")
        scriptMessageSubject.setMessage(u"Create winpe partition")
        logger.notice(u"Check winpe partition at number %s" % winpePartitionNumber)
        exitcode = os.system("test -b %s >> /dev/null 2>&1" % (getPartitionDevicePath(disk.device, winpePartitionNumber)))
        if (0 == int(exitcode)):
            logger.notice(u"%s exists" % (getPartitionDevicePath(disk.device, winpePartitionNumber)))
            partitionLabel = 'unknown'
            resultlist = execute('blkid -o value -s LABEL %s' % (getPartitionDevicePath(disk.device, winpePartitionNumber)))
            resultlist = list(filter(None, resultlist))
            if will_create_new_partition_table:
                disk.deletePartition (winpePartitionNumber)
            else:
                if useGpt:
                    execute('/usr/sbin/sgdisk -g --delete=%d -p %s' % (winpePartitionNumber, disk.device))
                    createPartitionEx("winpe", winpePartitionNumber, winpePartitionLabel, peStartM, winpePartitionSizeM, gptAutoAppend = True, boot = True, gptAttribute = "" , useGpt = useGpt, diskDevice = disk.device, fileSystem = "ntfs", gptFileSystem="0b00")
                    logger.notice(u"change id of partition %s to 0700" % getPartitionDevicePath(disk.device,winpePartitionNumber))
                    execute('/usr/sbin/sgdisk -t %d:0700 --attributes %d:clear:62 -p %s' % (winpePartitionNumber, winpePartitionNumber, disk.device))
                    execute('/usr/sbin/sgdisk --info=%d %s' % (winpePartitionNumber, disk.device))
                else:
                    execute('sfdisk --delete %s %s' % (disk.device, winpePartitionNumber))
                    createPartitionEx("winpe", winpePartitionNumber, winpePartitionLabel, peStartM, winpePartitionSizeM, gptAutoAppend = True, boot = True, gptAttribute = "" , useGpt = useGpt, diskDevice = disk.device, fileSystem = "ntfs", gptFileSystem="0b00")
                    execute('sfdisk --part-type %s %s 7' % (disk.device, winpePartitionNumber))
                createFilesystemEx("winpe", winpePartitionNumber, winpePartitionLabel, "ntfs", disk.device , useGpt)
            for resultline in resultlist:
                try:
                    partitionLabel = resultline.strip()
                except:
                    pass
            logger.notice(u"%s exists as device with label: %s" % (getPartitionDevicePath(disk.device, winpePartitionNumber), partitionLabel) )
            if (partitionLabel == winpePartitionLabel):
                logger.notice(u"%s exists as expected - reusing it" % (getPartitionDevicePath(disk.device, winpePartitionNumber)))
                logger.notice(u"Activate (set boot flag) on partition %s" % (getPartitionDevicePath(disk.device,winpePartitionNumber)))
                if inUefi:
                    deleteAllEntriesFromEFIBootmgr('opsitempwinpe')
                    logger.notice(u"setting opsitempwinpe to the top of uefi boot order")
                else:
                    execute('sfdisk -L --no-reread -A %s %d' % (disk.device,winpePartitionNumber), captureStderr=False)
            else:
                if (partitionLabel == dataPartitionLabel):
                    create_new_datapartition = False
                    if useGpt:
                        if isOpsi43:
                            show_error("Moving data partition is not implemented for GPT. Could not create winpe partition on %s - stopping here" % (disk.device))
                        raise Exception("Moving data partition is not implemented for GPT. Could not create winpe partition on %s - stopping here" % (disk.device))
                    else:
                        logger.notice(u"%s exists as expected for data partition - ok - moving partition %s one up and inserting new partition %s" % (getPartitionDevicePath(disk.device, winpePartitionNumber), winpePartitionNumber, winpePartitionNumber) )
                        partitions = disk.getPartitions()
                        if bootPartitionSize.startswith("0"):
                            winpePartitionSizeM = int(partitions[1]['secStart'] - partitions[0]['secEnd'] - 200)
                            logger.notice(u"New partition %s as start: %dS  size: %dS  type: vfat boot:true" % (winpePartitionNumber, int(partitions[0]['secEnd'] + 100), winpePartitionSizeM) )
                            cmd = u'%s -e "' % which('echo')
                            # let part1 as it is
                            cmd += u'%s,%s,%s\n' % (partitions[0]['secStart'], partitions[0]['secSize'], partitions[0]['type'])
                            # new  part 2
                            cmd += u'%d,%d,%s,*\n' % (int(partitions[0]['secEnd'] + 100), winpePartitionSizeM, 'b')
                            # move part 2 to 3
                            cmd += u'%s,%s,%s\n' % (partitions[1]['secStart'], partitions[1]['secSize'], partitions[1]['type'])
                            cmd +=  u'" | %s -L --no-reread -uS -f %s' % (which('sfdisk'), disk.device)
                        else:
                            winpePartitionSizeM = int(partitions[2]['secStart'] - partitions[1]['secEnd'] - 200)
                            logger.notice(u"New partition %s as start: %dS  size: %dS  type: vfat boot:true" % (winpePartitionNumber, int(partitions[0]['secEnd'] + 100), winpePartitionSizeM) )
                            cmd = u'%s -e "' % which('echo')
                            # let part1 as it is
                            cmd += u'%s,%s,%s\n' % (partitions[0]['secStart'], partitions[0]['secSize'], partitions[0]['type'])
                            # let part2 as it is
                            cmd += u'%s,%s,%s\n' % (partitions[1]['secStart'], partitions[1]['secSize'], partitions[1]['type'])
                            # new  part 3
                            cmd += u'%d,%d,%s,*\n' % (int(partitions[1]['secEnd'] + 100), winpePartitionSizeM, 'b')
                            # move part 3 to 4
                            cmd += u'%s,%s,%s\n' % (partitions[2]['secStart'], partitions[2]['secSize'], partitions[2]['type'])
                            cmd +=  u'" | %s -L --no-reread -uS -f %s' % (which('sfdisk'), disk.device)
                        execute(cmd, captureStderr=False)
                        execute('%s %s' % (which('partprobe'),disk.device))
                        time.sleep(2)
                        partitions = disk.getPartitions()
                else:
                    logger.notice(u"%s exists as unknown partition - giving up" % (getPartitionDevicePath(disk.device, winpePartitionNumber)))
                    if isOpsi43:
                        show_error("Could not create winpe partition on %s - stopping here" % (disk.device))
                    raise Exception("Could not create winpe partition on %s - stopping here" % (disk.device))
        else:
            if inUefi:
                createPartitionEx("winpe", winpePartitionNumber, winpePartitionLabel, peStartM, winpePartitionSizeM, gptAutoAppend = True, boot = True, gptAttribute = "" , useGpt = useGpt, diskDevice = disk.device, fileSystem = "fat32", gptFileSystem="0b00")
            else:
                createPartitionEx("winpe", winpePartitionNumber, winpePartitionLabel, peStartM, winpePartitionSizeM, gptAutoAppend = True, boot = True, gptAttribute = "" , useGpt = useGpt, diskDevice = disk.device, fileSystem = "ntfs", gptFileSystem="0b00")
            #createPartitionEx("winpe", winpePartitionNumber, winpePartitionLabel, peStartM, winpePartitionSizeM, gptAutoAppend = True, boot = True, gptAttribute = "" , useGpt = useGpt, diskDevice = disk.device, fileSystem = "ntfs", gptFileSystem="0700")

        if create_new_datapartition and (dataPartitionNumber != 0):
            # Create data partition
            logger.notice(u"Check data partition")
            scriptMessageSubject.setMessage(u"Create data partition")
            logger.notice(u"Check data partition at number %s" % dataPartitionNumber)
            exitcode = os.system("test -b %s >> /dev/null 2>&1" % (getPartitionDevicePath(disk.device, dataPartitionNumber)))
            if (0 == int(exitcode)):
                logger.notice(u"%s exists" % (getPartitionDevicePath(disk.device, dataPartitionNumber)))
                partitionLabel = 'unknown'
                resultlist = execute('blkid -o value -s LABEL %s' % (getPartitionDevicePath(disk.device, dataPartitionNumber)))
                resultlist = list(filter(None, resultlist))
                for resultline in resultlist:
                    try:
                        partitionLabel = resultline.strip()
                    except:
                        pass
                logger.notice(u"%s exists as device with label: %s" % (getPartitionDevicePath(disk.device, dataPartitionNumber), partitionLabel) )
                if (partitionLabel == dataPartitionLabel):
                    logger.notice(u"%s exists as expected - reusing it" % (getPartitionDevicePath(disk.device, dataPartitionNumber)))
                    create_new_datapartition = False

        if create_new_datapartition and (dataPartitionNumber != 0):
            createPartitionEx("data", dataPartitionNumber, dataPartitionLabel, int(peEndM), int(-1), gptAutoAppend = True, boot = False, gptAttribute = "" , useGpt = useGpt, diskDevice = disk.device, fileSystem = "fat32", gptFileSystem="0700")
        if not useGpt:
            partitions = disk.getPartitions()

        # Create ntfs filesystem on winpe partition
        testPartitionEx(disk.device, winpePartitionNumber, useGpt)
        checkClearPartition("none", disk.device, winpePartitionNumber)
        if inUefi:
            logger.notice(u"We use fat32 for winpe because we are on uefi")
            createFilesystemEx("winpe", winpePartitionNumber, winpePartitionLabel, "fat32", disk.device , useGpt)
        else:
            createFilesystemEx("winpe", winpePartitionNumber, winpePartitionLabel, "ntfs", disk.device , useGpt)
        #createFilesystemEx("winpe", winpePartitionNumber, winpePartitionLabel, "ntfs", disk.device , useGpt)

        if bootPartitionNumber and bootPartitionHidden:
            if inUefi:
                logger.notice(u"change id of partition %s%d to 0700 / 63" % (disk.device,bootPartitionNumber))
                execute('/usr/sbin/sgdisk -t %d:0700 --attributes %d:set:63 -p %s' % (bootPartitionNumber, bootPartitionNumber, disk.device))
                execute('/usr/sbin/sgdisk --info=%d %s' % (bootPartitionNumber, disk.device))
                logger.notice(u"change id of partition %s%d to 0b00 / " % (disk.device,winpePartitionNumber))
                execute('/usr/sbin/sgdisk -t %d:0b00 --attributes %d:clear:63 -p %s' % (winpePartitionNumber, winpePartitionNumber, disk.device))
                execute('/usr/sbin/sgdisk --info=%d %s' % (winpePartitionNumber, disk.device))
            else:
                logger.notice(u"change id of partition %s%d to 12" % (disk.device,bootPartitionNumber))
                execute('sfdisk --change-id %s %d 12' % (disk.device,bootPartitionNumber))
                logger.notice(u"change id of partition %s%d to 0b" % (disk.device,winpePartitionNumber))
                execute('sfdisk --change-id %s %d 0b' % (disk.device,winpePartitionNumber))
            execute('%s %s' % (which('partprobe'),disk.device))
            time.sleep(2)
            partitions = disk.getPartitions()
            execute('sfdisk -l %s' % (disk.device))

        if not useGpt:
            # copied from posix.px line2045
            retries = 1
            while retries <= 6:
                if os.path.exists(disk.getPartition(winpePartitionNumber)['device']):
                    break
                retries += 1
                if retries == 3:
                    logger.debug("Forcing kernel to reread the partitiontable again")
                    disk._forceReReadPartionTable()
                time.sleep(2)

        if create_new_datapartition and (dataPartitionNumber != 0):
            # Create ntfs filesystem on data partition
            createFilesystemEx("data", dataPartitionNumber, dataPartitionLabel,  fileSystem = "ntfs" , diskDevice = disk.device , useGpt = useGpt)

    if olimode:
        # System partition will be deleted => delete all netboot product states
        deletePocs = []
        for poc in backend.productOnClient_getObjects(clientId = clientId):
            if ((poc.productType == 'NetbootProduct') and (poc.productId != productId)) or \
               ((poc.productType == 'LocalbootProduct') and (poc.installationStatus == 'not_installed') and (poc.actionRequest == 'none')):
                deletePocs.append(poc)
        backend.productOnClient_deleteObjects(deletePocs)

        # Create ntfs filesystem on system partition
        testPartitionEx(disk.device, systemPartitionNumber, useGpt)
        checkClearPartition("none", disk.device, systemPartitionNumber)
        createFilesystemEx("system", systemPartitionNumber, "system", "ntfs", disk.device , useGpt)

        # Create ntfs filesystem on winpe partition
        testPartitionEx(disk.device, winpePartitionNumber, useGpt)
        checkClearPartition("none", disk.device, winpePartitionNumber)
        if inUefi:
            logger.notice(u"Create fat32 filesystem on winpe partition")
            createFilesystemEx("winpe", winpePartitionNumber, "winpe", "fat32", disk.device , useGpt)
        else:
            createFilesystemEx("winpe", winpePartitionNumber, "winpe", "ntfs", disk.device , useGpt)
        #createFilesystemEx("winpe", winpePartitionNumber, "winpe", "ntfs", disk.device , useGpt)
    if not useGpt:
        # Write Master Boot Record
        #disk.writeMasterBootRecord(system = 'vista')
        logger.notice(u"Write Master Boot Record")
        execute("/usr/local/bin/ms-sys -v")
        if int(bootimageVersion) < 20181022:
            execute("/usr/local/bin/ms-sys --mbrnt60 %s" % (disk.device))
        else:
            execute("/usr/local/bin/ms-sys --mbr7 %s" % (disk.device))
            # we need a sleep on nvme devices
            time.sleep(5)
            execute("/usr/local/bin/ms-sys --ntfs %s" % (getPartitionDevicePath(disk.device, winpePartitionNumber)))
    else:
        if inUefi:
            logger.notice(u"check efi boot order")
            execute('/bin/efibootmgr')

    execute('%s %s' % (which('partprobe'),disk.device))
    time.sleep(2)
    partitions = disk.getPartitions()
    execute('sfdisk -l %s' % (disk.device))
    if not useGpt:
        # copied from posix.px line2045
        retries = 1
        while retries <= 6:
            if os.path.exists(disk.getPartition(winpePartitionNumber)['device']):
                break
            retries += 1
            if retries == 3:
                logger.debug("Forcing kernel to reread the partitiontable again")
                disk._forceReReadPartionTable()
            time.sleep(2)

    #raw_input('Press <ENTER> to continue')

    # Create diskinfo.ini list
    logger.notice(u"collect disk and partition infos")
    diskuuid = getUuidFromDiskDevice(disk.device)
    logger.notice('we have uuid %s on %s' % (diskuuid,disk.device))
    if (diskuuid == '') or (diskuuid == 'None') or (not diskuuid):
        logger.notice('create new random disk uuid')
        import random
        exitcode = os.system('echo "x\ni\n%s\nr\nw"  | fdisk %s' % (random.randrange(0xFFFFFFFF),disk.device))
        if (0 != int(exitcode)):
            logger.notice(u"Failed to set uuid on disk: %s - retry" % disk.device)
        execute('%s %s' % (which('partprobe'),disk.device))
        time.sleep(2)
        #execute('echo "x\ni\n%s\nr\nw"  | fdisk %s' % (random.randrange(0xFFFFFFFF),disk.device))

        diskuuid = getUuidFromDiskDevice(disk.device)
        logger.notice('we have uuid %s on %s' % (diskuuid,disk.device))
        if (diskuuid == '') or (diskuuid == 'None') or (not diskuuid):
            logger.notice('create new random disk uuid')
            import random
            exitcode = os.system('echo "x\ni\n%s\nr\nw"  | fdisk %s' % (random.randrange(0xFFFFFFFF),disk.device))
            if (0 != int(exitcode)):
                logger.notice(u"Failed to set uuid on disk: %s - retry" % disk.device)
            execute('%s %s' % (which('partprobe'),disk.device))
            time.sleep(2)
            #execute('echo "x\ni\n%s\nr\nw"  | fdisk %s' % (random.randrange(0xFFFFFFFF),disk.device))

            diskuuid = getUuidFromDiskDevice(disk.device)
            logger.notice('we have uuid %s on %s' % (diskuuid,disk.device))
    execute('blkid')
    #execute('blkid -p -s PTUUID -o value %s' % disk.device)
    #execute('blkid -p -s PTUUID -o value %s' % disks[0].device)
    #execute('blkid -p -s UUID -o value %s' % getPartitionDevicePath(disks[0].device,1))
    #execute('udevadm info -q all -n  %s' % disks[0].device)


    allPartitionIdBackup = []
    allDiskBackup = []
    maxDiskRange = int(diskindex+1)
    numberOfDisks = len(disks)
    if (multiDiskMode == "0_ignore_anything_else"):
        diskindex = 0
        maxDiskRange = int(1)
        numberOfDisks = 1
    if useRaid1:
        diskindex = 0
        maxDiskRange = int(2)
    else:
        logger.notice(u"We save all partition infos ....")
        for i in range(numberOfDisks):
            aktdisk = disks[i]
            raidInfo = 'none'
            if isRaidMember(aktdisk.device):
                raidInfo = 'raidMember'
            if isRaidDisk(aktdisk.device):
                raidInfo = 'raidDisk'
            if raidInfo == 'raidMember':
                # Do not try to analyze Raid Member disks
                allDiskBackup.append([i,aktdisk.device,True,0,'',raidInfo])
            else:
                execute('sfdisk -l %s' % (aktdisk.device))
                logger.notice(u"has disk %s a gpt partition table ?" % (aktdisk.device))
                fdiskexitcode = os.system('blkid -s PTTYPE %s | grep -qF gpt' % (aktdisk.device))
                if (0 != int(fdiskexitcode)):
                    logger.notice(u"disk %s have no gpt partition table" % (aktdisk.device))
                    aktDiskIsGPT = False
                    numberOfPartitions = len(aktdisk.getPartitions())
                else:
                    logger.notice(u"disk %s have a gpt partition table:" % (aktdisk.device))
                    aktDiskIsGPT = True
                    output = execute('sgdisk -p %s' % (aktdisk.device))
                    output = list(filter(None, output))
                    output = output[len(output)-1].split()
                    if output[0] == 'Number':
                        numberOfPartitions = 0
                    else:
                        numberOfPartitions = int(output[0])
                logger.notice(u"We have %d partitions on disk %s" % (numberOfPartitions,aktdisk.device))
                diskuuid = getUuidFromDiskDevice(aktdisk.device)
                logger.notice(u"We have uuid %s  on disk %s" % (diskuuid,aktdisk.device))
                allDiskBackup.append([i,aktdisk.device,aktDiskIsGPT,numberOfPartitions,diskuuid,raidInfo])
                for aktpartnum in range(1,numberOfPartitions+1):
                    partuuid = getPartuuidFromPartitionDevice(getPartitionDevicePath(aktdisk.device,aktpartnum))
                    if aktDiskIsGPT:
                        output = execute('/usr/sbin/sgdisk --info=%d %s | grep "Attribute flags"' % (aktpartnum, aktdisk.device))
                        output = output[0].split(':')
                        oldPartitionId = output[1]
                        oldPartitionId = oldPartitionId.strip()
                    else:
                        output = execute('sfdisk --print-id %s %d' % (aktdisk.device,aktpartnum))
                        oldPartitionId = output[0].strip()
                    allPartitionIdBackup.append([i,aktpartnum,aktDiskIsGPT,oldPartitionId,partuuid])


    diskinfos = []
    diskinfos.append('[general]')
    diskinfos.append("disknumber=%d\r" % diskindex)
    diskinfos.append("installto=%d\r" % installdiskindex)
    diskinfos.append("diskcount=%d" % len(disks))
    for aktdisk in allDiskBackup:
        diskinfos.append("diskuuidbynum%d=%s" % (aktdisk[0],aktdisk[4]))
        diskinfos.append("disknumbyuuid%s=%d" % (aktdisk[4],aktdisk[0]))
    if diskmode:
        diskinfos.append("diskmode=disk")
    if vhdmode:
        diskinfos.append("diskmode=vhd")
    if olimode:
        diskinfos.append("diskmode=oli")
    for aktdisk in allDiskBackup:
        diskinfos.append('[%s]' % aktdisk[4])
        diskinfos.append('number=%s' % aktdisk[0])
        diskinfos.append('device=%s' % aktdisk[1])
        diskinfos.append('gpt=%s' % aktdisk[2])
        diskinfos.append('partitions=%s' % aktdisk[3])
        diskinfos.append('raidinfo=%s' % aktdisk[5])
        diskinfos.append('[%s]' % aktdisk[0])
        diskinfos.append('device=%s' % aktdisk[1])
        diskinfos.append('gpt=%s' % aktdisk[2])
        diskinfos.append('partitions=%s' % aktdisk[3])
        diskinfos.append('uuid=%s' % aktdisk[4])
        diskinfos.append('raidinfo=%s' % aktdisk[5])
    for aktpart in allPartitionIdBackup:
        diskinfos.append('[disk%d-part%d]' % (aktpart[0],aktpart[1]))
        diskinfos.append('gpt=%s' % aktpart[2])
        diskinfos.append('atr=%s' % aktpart[3])
        diskinfos.append('uid=%s' % getUuidFromPartitionDevice(getPartitionDevicePath(disks[aktpart[0]].device,aktpart[1])))
        diskinfos.append('pid=%s' % getPartuuidFromPartitionDevice(getPartitionDevicePath(disks[aktpart[0]].device,aktpart[1])))
        for aktoldpart in unusedPartitionIdBackup:
            if (aktoldpart[0] == aktpart[0]) and (aktoldpart[1] == aktpart[1]):
                diskinfos.append('atrold=%s' % aktoldpart[3])

    if isOpsi43:
        show_message("Mounting partitions and copying data", 40)
    # Mount partition
    logger.notice(u"mount winpe partition")
    checkClearMountpoint(target, disk.device, winpePartitionNumber)
    checkClearPartition(target, disk.device, winpePartitionNumber)
    if inUefi:
        logger.notice(u"We use vfat for winpe because we are on uefi")
        checkMount(target, disk.device, winpePartitionNumber, "vfat")
    else:
        checkMount(target, disk.device, winpePartitionNumber, "ntfs")
    if not inUefi:
        depot.copy(source + '/' + winpeDir + '/*', target)
        logger.notice(u"we do not copy the uefi boot files because we are not in uefi mode")
        if os.path.exists(target + '/EFI'):
            execute('rm -r %s/EFI' % target)
            #depot.delete(target + '/EFI')
        if os.path.exists(target + '/bootmgr.efi'):
            execute('rm %s/bootmgr.efi' % target)
        if os.path.exists(target + '/BOOTMGR.EFI'):
            execute('rm %s/BOOTMGR.EFI' % target)
        if not (os.path.exists(os.path.join(target, 'bootmgr')) or os.path.exists(os.path.join(target, 'BOOTMGR'))):
            if isOpsi43:
                show_message("bootmgr not found in winpe, please check winpe folder on server")
            raise Exception("bootmgr not found in winpe, please check winpe folder on server")
    else:
        # here is the win8.1 uefi winpe
        depot.copy(source + '/' + winpeDir + '/*', target)
        if not (os.path.exists(os.path.join(target, 'bootmgr.efi')) or os.path.exists(os.path.join(target, 'BOOTMGR.EFI'))):
            if isOpsi43:
                show_message("bootmgr.efi not found in winpe_uefi, please check winpe_uefi folder on server")
            raise Exception("bootmgr.efi not found in winpe_uefi, please check winpe_uefi folder on server")

    # Copy Windows installation files
    if not winpeNetworkMode:
        depot.copy(source + '/'+installfiles_dir, target)

    # Copy opsi files
    depot.copy(source + '/opsi', target + '/')

    if isOpsi43:
        show_message("Copy opsi-client-agent", 70)
    # Copy opsi-client-agent
    depot.copy('/opsi-client-agent', target + '/opsi/')
    copy(target + '/opsi/opsi-client-agent/files/opsi/postinst.d/*', target + '/opsi/postinst.d/')

    # Copy custom files
    depot.copy(source + '/custom/unattend.xml', target + '/opsi/')
    depot.copy(source + '/custom/postinst.d/*', target + '/opsi/postinst.d/')
    
    # Copy opsiscript files
    depot.copy(source + '/nt6_repartition.opsiscript', target + '/opsi/postinst.d/')
    depot.copy(source + '/10_dhcp.opsiscript', target + '/opsi/postinst.d/')

    customUnattendMd5 = md5(open(target + '/opsi/unattend.xml', 'rb').read()).hexdigest()
    if unattendTemplateMd5 != customUnattendMd5:
        logger.error("unattend.xml differs from unattend.xml.template")
        scriptMessageSubject.setMessage("unattend.xml differs from unattend.xml.template")

    # Copy sysconf.ini to opsi-client-agent config dir
    copy(SYSCONF_INI , target + '/opsi/opsi-client-agent/files/opsi/cfg')

    # Copy opsi-script interpreter
    if action_processor == "opsi-winst":
        if not depot.exists("/opsi-winst/files64/opsi-winst/winst64.exe"):
            if isOpsi43:
                show_message("Incomplete opsi-winst Packages (64 Bit Files missing) - please install opsi-winst >= 4.12.0.13-1")
            raise Exception("Incomplete opsi-winst Packages (64 Bit Files missing) - please install opsi-winst >= 4.12.0.13-1")
        depot.copy('/opsi-winst', target + '/opsi/opsi-script/')

    if action_processor == "opsi-script":
        if not depot.exists("/opsi-script/windows/x64/opsi-script.exe"):
            if isOpsi43:
                show_message("Incomplete opsi-script Package (64 Bit Files missing) - please install opsi-script >= 4.12.0.15-1")
            raise Exception("Incomplete opsi-script Package (64 Bit Files missing) - please install opsi-script >= 4.12.0.15-1")
        depot.copy('/opsi-script/windows', target + '/opsi/opsi-script/')
        depot.copy('/opsi-script/common/*', target + '/opsi/opsi-script/x86/')
        depot.copy('/opsi-script/common/*', target + '/opsi/opsi-script/x64/')

    # Copy opsi-script script
    depot.copy(source + '/setup.opsiscript', target + '/opsi/')
    if depot.exists("%s/%s.png" % (source,productId)):
        depot.copy(source + '/'+productId+'.png', target + '/opsi/')
    if depot.exists("%s/windows.png" % (source)):
        depot.copy(source + '/windows.png', target + '/opsi/')

    # Copy device manager
    # http://www.nirsoft.net/utils/device_manager_view.html
    depot.copy(source + '/opsi/32/DevManView.exe', target + '/opsi/32/')
    depot.copy(source + '/opsi/64/DevManView.exe', target + '/opsi/64/')

    if isOpsi43:
        show_message("Copy drivers", 80)
    # Integrate drivers
    os.makedirs(dstDriversDir)
    os.makedirs(dstPeDriversDir)
    logger.notice(u"installing drivers from %s" % srcDriversDir)
    if depot.exists(srcDriversDir):
        scriptMessageSubject.setMessage(u"Found driver directory: '%s' - start driver integration" % srcDriversDir)
        logger.notice(u"Found driver directory %s" % srcDriversDir)

        additionalDrivers = [ ad.strip() for ad in (','.join(productPropertyValues.get('additional_drivers', [u'']))).split(',') ]
        integrateAdditionalWindowsDrivers(srcDriversDir + u'/drivers/additional', dstDriversDir, additionalDrivers, messageSubject = scriptMessageSubject, srcRepository = depot, auditHardwareOnHosts = auditHardwareOnHosts)
        integrateWindowsHardwareDrivers(srcDriversDir, dstDriversDir, auditHardwareOnHosts, messageSubject = scriptMessageSubject, srcRepository = depot)

    # Create work.cmd
    logger.notice(u"Creating work.cmd")
    with open(target + '/opsi/work.cmd', 'w') as work:
        work.write("rem Version=4.1.0.0\r\n")
        work.write("echo 'bcdboot x:\\windows /v' >> x:\\work.log\r\n")
        work.write("bcdboot x:\\windows /v >> x:\\work.log 2>&1 \r\n")
        work.write("echo 'bcdedit /enum firmware' >> x:\\work.log\r\n")
        work.write("bcdedit /enum firmware >> x:\\work.log\r\n")
        work.write("@echo off\r\n")
        if winpe_debug_cmd_exe:
            work.write("start cmd.exe\r\n")
        work.write(":install\r\n")
        work.write("diskpart /s x:\\diskpartWinpe.txt >> x:\\work.log\r\n")
        work.write("w:\\opsi\\test32bit.exe >nul 2>nul\r\n")
        work.write("if ERRORLEVEL  1 goto run64\r\n")
        work.write(":run32\r\n")
        work.write("echo \"Running in 32 Bit PE\" >> x:\\work.log\r\n")
        work.write("echo \"Running in 32 Bit PE\"\r\n")
        work.write("echo Starting opsi-script ......>> x:\\work.log\r\n")
        if action_processor == "opsi-script":
            work.write("w:\\opsi\\opsi-script\\x86\\opsi-script.exe w:\\opsi\\setup.opsiscript /batch /logfile x:\\opsiscript.log\r\n")
        elif action_processor == "opsi-winst":
            work.write("w:\\opsi\\opsi-script\\files\\opsi-winst\\winst32.exe w:\\opsi\\setup.opsiscript /batch /logfile x:\\opsiscript.log\r\n")
        work.write("goto actionprocessorfinished\r\n")
        work.write(":run64\r\n")
        work.write("echo \"Running in 64 Bit PE\" >> x:\\work.log\r\n")
        work.write("echo \"Running in 64 Bit PE\"\r\n")
        work.write("x:\\windows\\system32\\bcdedit.exe /enum firmware >> x:\\work.log\r\n")
        work.write("echo Starting opsi-script ......>> x:\\work.log\r\n")
        if action_processor == "opsi-script":
            work.write("w:\\opsi\\opsi-script\\x64\\opsi-script.exe w:\\opsi\\setup.opsiscript /batch /logfile x:\\opsiscript.log\r\n")
        elif action_processor == "opsi-winst":
            work.write("w:\\opsi\\opsi-script\\files64\\opsi-winst\\winst64.exe w:\\opsi\\setup.opsiscript /batch /logfile x:\\opsiscript.log \r\n")
        work.write(":actionprocessorfinished\r\n")

    if winpeNetworkMode and (depot._url.startswith('smb:') or depot._url.startswith('cifs:')):
        match = re.search('^smb://([^/]+)/([^/]+)(.*)$', depot._url, re.IGNORECASE)
        if not match:
            if isOpsi43:
                show_error("Bad depot-URL '%s'" % depot._url)
            raise Exception("Bad depot-URL '%s'" % depot._url)
        hn = match.group(1)
        sn = match.group(2)
        pn = match.group(3)

        if os.path.exists(dstDriversDir):
            storageControllerInfo = disk.getControllerInfo()
        for filePath in findFiles(prefix = dstDriversDir, directory = dstDriversDir, includeFile = re.compile('\.inf$', re.IGNORECASE), returnDirs = False):
            infFile = InfFile(filePath)
            if infFile.isDeviceKnown(usedNetworkDevice['vendorId'], usedNetworkDevice['deviceId']):
                logger.info("Integrating winpe driver for network controller: %s" % usedNetworkDevice)
                filePath = filePath[len(target):]
                if filePath.startswith('/'):
                    filePath = filePath[1:]
                filePath = 'w:\\' + filePath.replace('/', '\\')
                logger.notice(u"Adding found driver %s to drvloadlist_org.txt\r" % filePath )
                with open(target + '/opsi/drvloadlist_org.txt', 'a') as drivers:
                    drivers.write("%s\r\n" % filePath)

            if (str(productProperties.get('storage_driver_in_drvpe', 'False')).lower() == 'true'):
                if storageControllerInfo and infFile.isDeviceKnown(storageControllerInfo['vendorId'], storageControllerInfo['deviceId']):
                    logger.info("Integrating winpe driver for storage controller: %s" % storageControllerInfo)
                    copy(os.path.dirname(filePath), dstPeDriversDir)
        if os.path.exists(target + '/opsi/drvloadlist_org.txt'):
            logger.notice(u"Here comes the %s/opsi/drvloadlist_org.txt" % target)
            execute('cat %s/opsi/drvloadlist_org.txt' % target)
        else:
            logger.notice(u"drvloadlist doesn't exist")

        encodedPcpatchPassword = list(backend.user_getCredentials(username=u'pcpatch', hostId = clientId).values())[0]
        shareusername = configStates.get('clientconfig.depot.user', ['%s\\pcpatch'% hn])[0]


    with open(target + '/opsi/diskpartWinpe.txt', 'w') as diskpart:
        diskpart.write("list disk\r\n")
        diskpart.write("select disk %d\r\n" % diskindex)
        diskpart.write("list partition\r\n")
        diskpart.write("list volume\r\n")
        diskpart.write("select volume C\r\n")
        diskpart.write("assign letter=%s\r\n" % winpePartitionLetter)
    logger.notice(u"Here comes the %s/opsi/diskpartWinpe.txt" % target)
    execute('cat %s/opsi/diskpartWinpe.txt' % target)

    if vhdmode:
        with open(target + '/opsi/diskpartvhd.txt', 'w') as diskpart:
            diskpart.write("list disk\r\n")
            diskpart.write("select disk %d\r\n" % diskindex)
            diskpart.write("list partition\r\n")
            diskpart.write("list volume\r\n")
            diskpart.write("select partition %d\r\n" % systemPartitionNumber)
            diskpart.write("assign letter=P\r\n")
            if useGpt:
                logger.notice(u"Do not hide vhd container partition in uefi: gptAttribute = 0")
                diskpart.write("gpt attributes=0x0000000000000000 \r\n")
            diskpart.write("create vdisk file=P:\WINDOWS.vhdx maximum=%d type=expandable\r\n" % windows_vhd_sizeM)
            diskpart.write("attach vdisk\r\n")
            if useGpt:
                # create leading efi / MS partions
                # https://docs.microsoft.com/en-us/previous-versions/windows/it-pro/windows-8.1-and-8/hh825686(v=win.10)
                diskpart.write("select disk %d\r\n" % (len(disks)))
                diskpart.write("clean\r\n")
                diskpart.write("convert gpt\r\n")
                diskpart.write("create partition efi size=400\r\n")
                diskpart.write('format quick fs=fat32 label="EFI"\r\n')
                diskpart.write("create partition msr size=128\r\n")
            diskpart.write("create partition primary\r\n")
            if not useGpt:
                diskpart.write("active\r\n")
            diskpart.write('format fs=ntfs quick label="WINDOWS"\r\n')
            diskpart.write("assign letter=R\r\n")
            diskpart.write("list volume\r\n")
            diskpart.write("list disk\r\n")
            diskpart.write("select disk %d\r\n" % (len(disks)))
            diskpart.write("list partition\r\n")
            diskpart.write("exit\r\n")

        logger.notice(u"Here comes the %s/opsi/diskpartvhd.txt" % target)
        execute('cat %s/opsi/diskpartvhd.txt' % target)

    # Create startnet.cmd
    logger.notice(u"Creating startnet.cmd")
    with open(target + '/opsi/startnet.cmd', 'w') as startnetcmd:
        startnetcmd.write("@echo off\r\n")
        if vhdmode:
            startnetcmd.write("copy c:\\opsi\\diskpartvhd.txt x:\\diskpartvhd.txt >nul\r\n")
        startnetcmd.write("copy c:\\opsi\\work.cmd x:\\work.cmd >nul\r\n")
        startnetcmd.write("copy c:\\opsi\\diskpartWinpe.txt x:\\diskpartWinpe.txt >nul\r\n")
        startnetcmd.write("x:\\work.cmd")

    # Create diskpart.txt
    with open(target + '/opsi/diskpart.txt', 'w') as diskpart:
        if preserve_winpe_partition:
            logger.notice(u"Creating diskpart.txt to preserve winpe")
            diskpart.write("list disk\r\n")
            diskpart.write("select disk %d\r\n" % diskindex)
            diskpart.write("list partition\r\n")
            diskpart.write("list volume\r\n")
            diskpart.write("select volume %s\r\n" % winpePartitionLetter)
            if useGpt:
                diskpart.write("gpt attributes=0x4000000000000000 \r\n")
            else:
                diskpart.write("set id=12 override noerr\r\n")
            if bootPartitionNumber and bootPartitionHidden:
                diskpart.write("select volume %s\r\n" % bootPartitionLetter)
                diskpart.write("remove letter=%s\r\n" % bootPartitionLetter)
            if dataPartitionCreate and (windowsPartitionSize != "100%"):
                diskpart.write("select volume %s\r\n" % dataPartitionLetter)
                if useGpt:
                    diskpart.write("gpt attributes=0x0000000000000000 \r\n")
                else:
                    diskpart.write("set id=7 override noerr\r\n")
            diskpart.write("list volume\r\n")
        else:
            logger.notice(u"Creating diskpart.txt to delete winpe")
            diskpart.write("list disk\r\n")
            diskpart.write("select disk %d\r\n" % diskindex)
            diskpart.write("list partition\r\n")
            diskpart.write("list volume\r\n")
            if dataPartitionCreate and (dataPartitionNumber != 0):
                diskpart.write("select volume %s\r\n" % (winpePartitionLetter))
                if useGpt:
                    diskpart.write("gpt attributes=0x0000000000000000 \r\n")
                else:
                    diskpart.write("set id=7 override noerr\r\n")
            diskpart.write("select volume %s\r\n" % winpePartitionLetter)
            diskpart.write("delete volume noerr override\r\n")
            # do not extend if (dataPartitionCreate=false) and (windowsPartitionSize != "100%")
            # because then the system partition will extend to the complete disk
            if dataPartitionCreate or (windowsPartitionSize == "100%"):
                diskpart.write("select volume c\r\n")
                diskpart.write("extend\r\n")
            if bootPartitionNumber and bootPartitionHidden:
                diskpart.write("select volume %s\r\n" % bootPartitionLetter)
                diskpart.write("remove letter=%s\r\n" % bootPartitionLetter)
        diskpart.write("list volume\r\n")
        diskpart.write("exit\r\n")
    # create diskpart file to get recivery partition number, if it exists (24h2)
    logger.notice("diskpartRecovery.txt created if in uefi mode")
    if inUefi:
        logger.notice("creating diskpartReovery.txt as we are in uefi mode")
        with open(target + '/opsi/diskpartRecovery.txt', 'w') as diskpartRecovery:
            diskpartRecovery.write("list disk\r\n")
            diskpartRecovery.write("select disk %d\r\n" % diskindex)
            diskpartRecovery.write("list partition\r\n")
            diskpartRecovery.write("exit\r\n")
            execute('cat %s/opsi/diskpartRecovery.txt' % target)

    # Create partunhide.txt
    if os.path.exists(target + '/opsi/partunhide.txt'):
        logger.warning("The partition backup file: %s - already exists - no override.." % (target + '/opsi/partunhide.txt') )
    else:
        if unusedPartitionIdBackup:
            logger.notice(u"write old partitionIds to %s/opsi/partunhide.txt" % target)
            with open(target + '/opsi/partunhide.txt', 'w') as partunhide:
                for aktpart in unusedPartitionIdBackup:
                    partunhide.write("select disk %d\r\n" % aktpart[0])
                    partunhide.write("select partition %d\r\n" % aktpart[1])
                    if aktpart[2]:
                        partunhide.write("gpt attributes=0x%s \r\n" % aktpart[3])
                    else:
                        partunhide.write("set id=%s override noerr\r\n" % aktpart[3])

                partunhide.write("list volume\r\n")
                partunhide.write("exit\r\n")

        else:
            logger.notice(u"No Partition IDs to backup - do not create %s/opsi/partunhide.txt" % target)

    # Create parthide.txt
    if os.path.exists(target + '/opsi/parthide.txt'):
        logger.warning("The partition hide file: %s - already exists - no override.." % (target + '/opsi/parthide.txt') )
    else:
        if unusedPartitionIdBackup:
            logger.notice(u"write hide partitionIds to %s/opsi/parthide.txt" % target)
            with open(target + '/opsi/parthide.txt', 'w') as parthide:
                for aktpart in unusedPartitionIdBackup:
                    parthide.write("select disk %d\r\n" % aktpart[0])
                    parthide.write("select partition %d\r\n" % aktpart[1])
                    if aktpart[2]:
                        parthide.write("gpt attributes=0x4000000000000000 \r\n")
                    else:
                        parthide.write("set id=12 override noerr\r\n")

                parthide.write("list volume\r\n")
                parthide.write("exit\r\n")

        else:
            logger.notice(u"No Partition IDs to backup - do not create %s/opsi/parthide.txt" % target)

    # Create diskinfo.ini
    if os.path.exists(target + '/opsi/diskinfo.ini'):
        logger.warning("The boot disk file: %s - already exists - no override.." % (target + '/opsi/diskinfo.ini') )
    else:
        logger.notice(u"write disk and partition infos to %s/opsi/diskinfo.ini" % target)
        with open(target + '/opsi/diskinfo.ini', 'w') as di:
            for line in diskinfos:
                di.write("%s\r\n" % line)

    logger.notice(u"Here comes the %s/opsi/diskinfo.ini" % target)
    execute('cat %s/opsi/diskinfo.ini' % target)
    #execute('blkid')
    #execute('blkid -p -s PTUUID -o value %s' % disk.device)
    #execute('blkid -p -s PTUUID -o value %s' % disks[0].device)
    #execute('blkid -p -s UUID -o value %s' % getPartitionDevicePath(disks[0].device,1))

    mac_address = ""
    if "enx" in usedNetworkDevice:
        mac_address = execute("ip link | grep -A1 enx | awk '{print $2}' | tail -n1")
  

    # Create opsi-script-infos.ini
    logger.notice(u"Creating opsi-script-infos.ini")
    with open(target + '/opsi/opsi-script-infos.ini', 'w') as osi:
        osi.write("[infos]\r\n")
        osi.write("clientId=%s\r\n" % clientId)
        osi.write("serviceurl=%s\r\n" % nextserver)
        osi.write("encodedPcpatchPassword=%s\r\n" % encodedPcpatchPassword)
        osi.write("shareusername=%s\r\n" % shareusername)
        osi.write("shareHostName=%s\r\n" % hn)
        osi.write("shareName=%s\r\n" % sn)
        osi.write("ProductId=%s\r\n" % productId)
        osi.write("diskindex=%s\r\n" % diskindex)
        osi.write("multiDiskMode=%s\r\n" % multiDiskMode)
        osi.write("installfiles_dir=%s\r\n" % installfiles_dir)
        osi.write("vhd_install=%s\r\n" % vhdmode)
        osi.write("diskcount=%s\r\n" % len(disks))
        osi.write("preservepe=%s\r\n" % preserve_winpe_partition)
        osi.write("winpeNetworkMode=%s\r\n" % winpeNetworkMode)
        osi.write("force_del_winpe_boot_entry=%s\r\n" % force_del_winpe_boot_entry)
        osi.write("del_winpe_boot_entry_from_boot_sequence=%s\r\n" % del_winpe_boot_entry_from_boot_sequence)
        if mac_address:
            osi.write("mac_address=%s\r\n" % mac_address[0])
        else:
            osi.write("mac_address=\r\n")

    if isOpsi43:
        show_message("patch files", 90)
    # Patch files
    logger.notice(u"Patching unattend.xml and config.ini")
    for i in ( target + '/opsi/unattend.xml', target + '/opsi/opsi-client-agent/files/opsi/cfg/config.ini', target + '/opsi/setup.opsiscript'):
        waitForFile(i)
        waitForFile(PATCHA_IN)
        execute('/usr/local/bin/patcha %s' % i)
        execute('/usr/local/bin/patcha -f %s %s' % (PATCHA_IN, i))
    logger.notice(u"Here comes the unattend.xml ")

    execute('cat %s/opsi/unattend.xml' % target)

    # read config.ini
    iniFile =  target + '/opsi/opsi-client-agent/files/opsi/cfg/config.ini'
    if not os.path.exists(iniFile):
        logger.warning("Could not find config.ini file: %s - bad.." % (iniFile) )
    else:
        # Wenn Datei existiert parse Datei und gebe ConfigParser Objekt zurueck
        iniFileHandler = IniFile(iniFile)
        config = iniFileHandler.parse()
        if not config.has_section("opsiclientd"):
            logger.warning("Could not find section opsiclientd in config.ini file.: - bad..")
        else:
            # read data
            systemurl = config.get("opsiclientd","config_service.url")
            systemurl = systemurl.strip()
            logger.notice(u"Reading config.ini data: config_service.url: %s " % (systemurl) )
            logger.notice(u" config_service.url from clientconfig: %s " % (nextserver) )
            if (systemurl != nextserver):
                logger.warning(" service url from config.ini: %s is different to clientconfig.configserver.url (commandline or input mask): %s" % (systemurl, nextserver))
                logger.notice(u"Using config.ini data as service url: config_service.url: %s " % (systemurl) )
    iniFileHandler.generate(config)

    # Debug output of file content
    for fn in ('/opsi/work.cmd', '/opsi/startnet.cmd', '/opsi/diskpart.txt', '/opsi/partunhide.txt'):
        if os.path.exists(target + fn):
            execute("%s '%s'" % (which('cat'), target + fn))

    if inUefi:
        currentbootentry = uefiGetActBootEntry()
        if currentbootentry:
            uefiNetbootLabel = currentbootentry['label']
            logger.notice('Current (net) boot entry: %s' % uefiNetbootLabel)
            backend.config_createUnicode("clientConfig.uefiNetbootLabel","clientspecific uefi firmware netboot entry label")
            backend.configState_create("clientConfig.uefiNetbootLabel", clientId, [ uefiNetbootLabel ] )
        else:
            logger.notice('No usable current (net) boot entry found')

        deleteAllEntriesFromEFIBootmgr('opsitempwinpe')
        deleteAllEntriesFromEFIBootmgr('winpe')
        deleteAllEntriesFromEFIBootmgr('ubuntu')
        logger.notice('deleting opsibootimage boot label')
        deleteAllEntriesFromEFIBootmgr('opsibootimage')
        deleteAllEntriesFromEFIBootmgr('installer')
        logger.notice(u"setting winpe to the top of winpe boot order")
        # This was a test for working with ntfs in winpe also in uefi
        # It did not work because we need win to create the bcd store on the efi partition
        # https://docs.microsoft.com/en-us/previous-versions/windows/it-pro/windows-7/hh265131(v=ws.10)
        ## create mount point for efi
        #efimount = '/efimount'
        #checkCreateDir(efimount)
        #checkClearMountpoint(efimount, disk.device, 1)
        #checkClearPartition(efimount, disk.device, 1)
        #checkMount(efimount, disk.device, 1, "vfat")
        #checkCreateDir(efimount+'/EFI/BOOT')
        ## https://docs.microsoft.com/en-us/previous-versions/windows/it-pro/windows-7/hh265131(v=ws.10)
        #copy(target + '/EFI/*', efimount + '/EFI')
        #copy(target + '/windows/boot/EFI/*.efi', efimount + '/EFI/Microsoft/Boot/')
        try:
            invalidArgument = False
            execute('/bin/efibootmgr -c --disk %s --part %s -L opsitempwinpe --loader \\\\EFI\\\\boot\\\\bootx64.efi' % (disk.device, winpePartitionNumber))
        except Exception as efibootmgrError:
            logger.error("efibootmgrError")
            logger.error("%s" % efibootmgrError)
            if "Invalid argument" in efibootmgrError.args[0]:
                logger.error("Something went wrong with efibootmgr 'Invalid argument'. Working around.")
                invalidArgument = True
                pass
        if not invalidArgument:
            setFirstUefiBoot('opsitempwinpe')
            setNextUefiBoot('opsitempwinpe')
        execute('/bin/efibootmgr -v')


    # Umount partition
    if useGpt:
        logger.notice(u"umount %s >> /tmp/gpt.log 2>&1" % (getPartitionDevicePath(disk.device, winpePartitionNumber)))
        execute('/bin/umount %s' % (getPartitionDevicePath(disk.device, winpePartitionNumber)))
    else:
        disk.umountPartition(partition = winpePartitionNumber)

    logger.notice(u"diskcounter is still %d" % (diskcounter))
# end of: for diskcounter in range(0,maxdiskindex)

handle_win_setupAfterInstall()


if olimode:
    # write log to data partition
    imageDir = '/mntimg'
    filesystem = 'ntfs'

    # mount the image (backup) file system
    # check the mount point
    disk = disks[backupdiskindex]
    checkCreateDir(imageDir)
    try:
        testPartitionEx(disk.device, backupPartitionNumber, useGpt)
    except:
        if isOpsi43:
            show_error("Could not find partition %d. - Please restart this product (opsi-local-image-prepare)" % backupPartitionNumber)
        raise Exception("Could not find partition %d. - Please restart this product (opsi-local-image-prepare)" % backupPartitionNumber)

    partitionLabel = partitionHasLabel(disk.device, backupPartitionNumber)
    if (partitionLabel.lower() == backupPartitionLabel.lower()):
        logger.notice(u"%s%s exists as expected - fine" % (disk.device, backupPartitionNumber) )
        logger.notice(u"%s%s created and formated" % (disk.device, backupPartitionNumber) )
    else:
        logger.error("%s%s exists not as expected with label %s - not ok - giving up" % (disk.device, backupPartitionNumber, backupPartitionLabel) )
        scriptMessageSubject.setMessage(u"%s%s exists not as expected with label %s - not ok - giving up" % (disk.device, backupPartitionNumber, backupPartitionLabel) )
        if isOpsi43:
            show_error("Wrong partition table - Please run the opsi product 'opsi-local-image-prepare' to prepare this disk ! ")
        raise Exception("Wrong partition table - Please run the opsi product 'opsi-local-image-prepare' to prepare this disk ! ")

    # check if mount point is in use
    checkClearMountpoint(imageDir, disk.device, backupPartitionNumber)
    # check if backup partition is in use
    checkClearPartition(imageDir, disk.device, backupPartitionNumber)
    # mount backup partition to mount point
    checkMount(imageDir, disk.device, backupPartitionNumber)

    if backupAfterInstall:
        logger.notice(u"Setting opsi-local-image-backup-starter to once")
        # set opsi-local-image-backup-starter to once
        backend.setProductActionRequest(productId='opsi-local-image-backup-starter', clientId=clientId, actionRequest='once')
        # Setting defaults for opsi-local-image-restore
        logger.notice(u"Setting defaults for opsi-local-image-restore to empty")
        imageFilesList = []
        backend.productPropertyState_create(productId =  'opsi-local-image-backup',propertyId = 'imagefile',objectId = clientId,values = imageFilesList)

    #if setupAfterInstall:
    #       for setupProduct in setupAfterInstall:
    #               setupProduct = setupProduct.strip()
    #               if setupProduct != '':
    #                       logger.notice(u"Set to setup %s" % (setupProduct) )
    #                       backend.setProductActionRequestWithDependencies(productId=setupProduct, clientId=clientId, actionRequest='setup')

    # write meta info to masterlog
    logstr = time.asctime( time.localtime(time.time()) )
    logstr = logstr + ', product=%s, filesystem=%s' % (productId, filesystem)
    logFile = imageDir+'/master.log'
    if not os.path.exists(logFile):
        f = open(logFile, 'w' )
    else:
        f = open(logFile, 'a' )
    f.write(logstr)
    f.close()
    # show
    logger.notice(u"Here comes the master log:")
    execute('cat %s' % (logFile) )

#exit(1)
# Reboot
reboot()
